from dataclasses import dataclass
import json
from typing import Callable

from proxyscope.app.config.runtime import RuntimeConfig, normalize_http_method, normalize_modification_url


@dataclass(frozen=True)
class CommandExecutionResult:
    handled: bool
    status_message: str = ""
    updated_log_level: int | None = None


class RuntimeCommandService:
    """
    Parse and execute runtime commands that mutate shared runtime configuration.
    """

    def __init__(self, *, runtime_config: RuntimeConfig) -> None:
        self._runtime_config = runtime_config

    def execute(
        self,
        command: str,
        *,
        on_cache_toggle: Callable[[], None] | None,
        on_schedule_policy_edit: Callable[[str], None],
    ) -> CommandExecutionResult:
        normalized = command.strip()
        if not normalized:
            return CommandExecutionResult(handled=False)

        parts = normalized.split()
        cmd = parts[0].lower()

        if cmd == "loglevel":
            return self._handle_loglevel(parts)
        if cmd in {"whitelist", "wl"}:
            return self._handle_whitelist(parts)
        if cmd == "cache":
            return self._handle_cache(parts, on_cache_toggle=on_cache_toggle)
        if cmd == "config":
            return self._handle_config(parts, on_cache_toggle=on_cache_toggle)
        if cmd in {"policy", "pol"}:
            return self._handle_policy(parts, on_schedule_policy_edit=on_schedule_policy_edit)
        return CommandExecutionResult(handled=False)

    def _handle_loglevel(self, parts: list[str]) -> CommandExecutionResult:
        if len(parts) == 1:
            return CommandExecutionResult(
                handled=True,
                status_message=f"Current log level: {self._runtime_config.log_level_name()}",
            )
        try:
            new_level = self._runtime_config.set_log_level(parts[1])
        except ValueError as exc:
            return CommandExecutionResult(handled=True, status_message=str(exc))
        return CommandExecutionResult(
            handled=True,
            status_message=f"Log level set to {self._runtime_config.log_level_name()}",
            updated_log_level=new_level,
        )

    def _handle_whitelist(self, parts: list[str]) -> CommandExecutionResult:
        if len(parts) == 1 or parts[1].lower() == "show":
            entries = self._runtime_config.whitelist_entries()
            if not entries:
                return CommandExecutionResult(
                    handled=True,
                    status_message="Whitelist is empty (logging enabled for all hosts).",
                )
            return CommandExecutionResult(
                handled=True,
                status_message="Whitelist: " + ", ".join(entries),
            )

        action = parts[1].lower()
        value = " ".join(parts[2:]).strip()

        if action == "add":
            if not value:
                return CommandExecutionResult(
                    handled=True,
                    status_message="Usage: whitelist add <host-or-url>",
                )
            try:
                added = self._runtime_config.add_whitelist_entry(value)
            except ValueError as exc:
                return CommandExecutionResult(handled=True, status_message=str(exc))
            return CommandExecutionResult(handled=True, status_message=f"Added to whitelist: {added}")

        if action == "remove":
            if not value:
                return CommandExecutionResult(
                    handled=True,
                    status_message="Usage: whitelist remove <host-or-url>",
                )
            try:
                removed = self._runtime_config.remove_whitelist_entry(value)
            except ValueError as exc:
                return CommandExecutionResult(handled=True, status_message=str(exc))
            if removed:
                return CommandExecutionResult(
                    handled=True,
                    status_message=f"Removed from whitelist: {value}",
                )
            return CommandExecutionResult(handled=True, status_message=f"Not in whitelist: {value}")

        if action == "clear":
            self._runtime_config.clear_whitelist()
            return CommandExecutionResult(
                handled=True,
                status_message="Whitelist cleared (logging enabled for all hosts).",
            )

        return CommandExecutionResult(
            handled=True,
            status_message="Usage: whitelist [show|add|remove|clear] ...",
        )

    def _handle_cache(
        self,
        parts: list[str],
        *,
        on_cache_toggle: Callable[[], None] | None,
    ) -> CommandExecutionResult:
        if len(parts) == 1 or parts[1].lower() == "show":
            enabled = self._runtime_config.cache_invalidation_enabled
            state = "on" if enabled else "off"
            return CommandExecutionResult(handled=True, status_message=f"Cache invalidation: {state}")

        action = parts[1].lower()
        if action == "on":
            self._runtime_config.set_cache_invalidation_enabled(True)
            _trigger_callback(on_cache_toggle)
            return CommandExecutionResult(handled=True, status_message="Cache invalidation enabled.")
        if action == "off":
            self._runtime_config.set_cache_invalidation_enabled(False)
            _trigger_callback(on_cache_toggle)
            return CommandExecutionResult(handled=True, status_message="Cache invalidation disabled.")
        if action == "toggle":
            enabled = self._runtime_config.toggle_cache_invalidation()
            _trigger_callback(on_cache_toggle)
            state = "enabled" if enabled else "disabled"
            return CommandExecutionResult(handled=True, status_message=f"Cache invalidation {state}.")

        return CommandExecutionResult(
            handled=True,
            status_message="Usage: cache [show|on|off|toggle]",
        )

    def _handle_policy(
        self,
        parts: list[str],
        *,
        on_schedule_policy_edit: Callable[[str], None],
    ) -> CommandExecutionResult:
        if len(parts) == 1 or parts[1].lower() == "show":
            entries = self._runtime_config.policy_descriptions()
            if not entries:
                return CommandExecutionResult(handled=True, status_message="No policy rules configured.")
            preview = ", ".join(entries[:3])
            if len(entries) > 3:
                preview += ", ..."
            return CommandExecutionResult(
                handled=True,
                status_message=f"Policies ({len(entries)}): {preview}",
            )

        action = parts[1].lower()
        if action == "add-static":
            parsed = _parse_add_static_policy_arguments(parts[2:])
            if parsed is None:
                return CommandExecutionResult(
                    handled=True,
                    status_message="Usage: policy add-static [METHOD] <url> [status] [content-type] [body...]",
                )
            method, url, status_code, content_type, body_text = parsed
            if status_code < 100 or status_code > 599:
                return CommandExecutionResult(
                    handled=True,
                    status_message="Status code must be in range 100..599.",
                )
            headers = {"Content-Type": content_type}
            rule_name = self._runtime_config.add_static_response_rule(
                url=url,
                status_code=status_code,
                reason=_default_reason_phrase(status_code),
                headers=headers,
                body=body_text.encode("utf-8"),
                method=method,
                url_prefix=False,
            )
            return CommandExecutionResult(handled=True, status_message=f"Added static policy: {rule_name}")

        if action == "add-editor":
            method, value = _parse_modify_method_and_url(parts[2:])
            if not value:
                return CommandExecutionResult(
                    handled=True,
                    status_message="Usage: policy add-editor [METHOD] <url>",
                )
            try:
                normalized = self._runtime_config.add_open_editor_policy(value, method=method or "GET")
            except ValueError as exc:
                return CommandExecutionResult(handled=True, status_message=str(exc))
            return CommandExecutionResult(handled=True, status_message=f"Added editor policy: {normalized}")

        if action == "remove-editor":
            method, value = _parse_modify_method_and_url(parts[2:])
            if not value:
                return CommandExecutionResult(
                    handled=True,
                    status_message="Usage: policy remove-editor [METHOD] <url>",
                )
            try:
                removed = self._runtime_config.remove_open_editor_policy(value, method=method)
                normalized = normalize_modification_url(value)
                method_label = normalize_http_method(method) if method is not None else "*"
            except ValueError as exc:
                return CommandExecutionResult(handled=True, status_message=str(exc))
            if removed:
                return CommandExecutionResult(
                    handled=True,
                    status_message=f"Removed editor policy: {method_label} {normalized}",
                )
            return CommandExecutionResult(
                handled=True,
                status_message=f"Editor policy not found: {method_label} {normalized}",
            )

        if action == "clear-editor":
            self._runtime_config.clear_open_editor_policies()
            return CommandExecutionResult(
                handled=True,
                status_message="Cleared all open-editor policies.",
            )

        if action == "remove":
            name = " ".join(parts[2:]).strip()
            if not name:
                return CommandExecutionResult(handled=True, status_message="Usage: policy remove <name>")
            try:
                removed = self._runtime_config.remove_policy_rule(name)
            except ValueError as exc:
                return CommandExecutionResult(handled=True, status_message=str(exc))
            if removed:
                return CommandExecutionResult(handled=True, status_message=f"Removed policy: {name}")
            return CommandExecutionResult(handled=True, status_message=f"Policy not found: {name}")

        if action in {"enable", "disable"}:
            name = " ".join(parts[2:]).strip()
            if not name:
                return CommandExecutionResult(
                    handled=True,
                    status_message=f"Usage: policy {action} <name>",
                )
            try:
                changed = self._runtime_config.set_policy_rule_enabled(name, enabled=(action == "enable"))
            except ValueError as exc:
                return CommandExecutionResult(handled=True, status_message=str(exc))
            if changed:
                state = "enabled" if action == "enable" else "disabled"
                return CommandExecutionResult(handled=True, status_message=f"Policy {state}: {name}")
            return CommandExecutionResult(handled=True, status_message=f"Policy not found: {name}")

        if action == "edit":
            name = " ".join(parts[2:]).strip()
            if not name:
                return CommandExecutionResult(handled=True, status_message="Usage: policy edit <name>")
            try:
                rule = self._runtime_config.get_policy_rule(name)
            except ValueError as exc:
                return CommandExecutionResult(handled=True, status_message=str(exc))
            if rule is None:
                return CommandExecutionResult(handled=True, status_message=f"Policy not found: {name}")
            on_schedule_policy_edit(name)
            return CommandExecutionResult(handled=True, status_message=f"Opening policy editor: {name}")

        return CommandExecutionResult(
            handled=True,
            status_message="Usage: policy [show|add-editor|remove-editor|clear-editor|add-static|edit|remove|enable|disable] ...",
        )

    def _handle_config(
        self,
        parts: list[str],
        *,
        on_cache_toggle: Callable[[], None] | None,
    ) -> CommandExecutionResult:
        if len(parts) == 1 or parts[1].lower() == "show":
            path = self._runtime_config.config_path
            if path is None:
                return CommandExecutionResult(handled=True, status_message="Config path: <not attached>")
            return CommandExecutionResult(handled=True, status_message=f"Config path: {path}")

        action = parts[1].lower()
        if action == "save":
            target = " ".join(parts[2:]).strip()
            try:
                if target:
                    saved_path = self._runtime_config.save_to_path(target)
                else:
                    current = self._runtime_config.config_path
                    if current is None:
                        return CommandExecutionResult(
                            handled=True,
                            status_message="Usage: config save <path> (or attach --config at startup)",
                        )
                    self._runtime_config.save()
                    saved_path = current
            except (OSError, ValueError) as exc:
                return CommandExecutionResult(handled=True, status_message=f"Config save failed: {exc}")
            return CommandExecutionResult(handled=True, status_message=f"Config saved: {saved_path}")

        if action == "reload":
            before_cache = self._runtime_config.cache_invalidation_enabled
            try:
                reloaded = self._runtime_config.reload_from_attached_file()
            except (OSError, ValueError, json.JSONDecodeError) as exc:
                return CommandExecutionResult(handled=True, status_message=f"Config reload failed: {exc}")
            if not reloaded:
                return CommandExecutionResult(
                    handled=True,
                    status_message="No attached config path. Use: config save <path>",
                )
            if before_cache != self._runtime_config.cache_invalidation_enabled:
                _trigger_callback(on_cache_toggle)
            return CommandExecutionResult(
                handled=True,
                status_message=f"Config reloaded: {self._runtime_config.config_path}",
                updated_log_level=self._runtime_config.log_level,
            )

        return CommandExecutionResult(
            handled=True,
            status_message="Usage: config [show|save [path]|reload]",
        )


def _trigger_callback(callback: Callable[[], None] | None) -> None:
    if callback is None:
        return
    callback()


def _parse_modify_method_and_url(tokens: list[str]) -> tuple[str | None, str]:
    if not tokens:
        return None, ""
    if len(tokens) == 1:
        return None, tokens[0]
    first = tokens[0].strip()
    if not first:
        return None, " ".join(tokens[1:]).strip()
    known_methods = {
        "GET",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
        "HEAD",
        "OPTIONS",
        "TRACE",
        "CONNECT",
    }
    if first.upper() in known_methods:
        return first.upper(), " ".join(tokens[1:]).strip()
    return None, " ".join(tokens).strip()


def _parse_add_static_policy_arguments(tokens: list[str]) -> tuple[str, str, int, str, str] | None:
    method, value = _parse_modify_method_and_url(tokens)
    method_value = method or "GET"
    if not value:
        return None

    parsed_tokens = value.split()
    if not parsed_tokens:
        return None
    url = parsed_tokens[0]
    status_code = 200
    content_type = "text/plain; charset=utf-8"
    body_text = ""

    offset = 1
    if len(parsed_tokens) > offset and parsed_tokens[offset].isdigit():
        status_code = int(parsed_tokens[offset])
        offset += 1
    if len(parsed_tokens) > offset:
        content_type = parsed_tokens[offset]
        offset += 1
    if len(parsed_tokens) > offset:
        body_text = " ".join(parsed_tokens[offset:])

    return method_value, url, status_code, content_type, body_text


def _default_reason_phrase(status_code: int) -> str:
    mapping = {
        200: "OK",
        201: "Created",
        202: "Accepted",
        204: "No Content",
        301: "Moved Permanently",
        302: "Found",
        400: "Bad Request",
        401: "Unauthorized",
        403: "Forbidden",
        404: "Not Found",
        409: "Conflict",
        418: "I'm a teapot",
        429: "Too Many Requests",
        500: "Internal Server Error",
        502: "Bad Gateway",
        503: "Service Unavailable",
        504: "Gateway Timeout",
    }
    return mapping.get(status_code, "OK")



from dataclasses import dataclass
import gzip
import os
from pathlib import Path
import shlex
import shutil
import subprocess
import tempfile
import zlib

from proxyscope.app.config.runtime import get_runtime_config
from proxyscope.app.editing.modifier import PendingResponseEdit


@dataclass(frozen=True)
class _BodyEditPlan:
    is_text: bool
    charset: str
    editor_body: bytes
    remove_content_encoding_header: bool


def edit_pending_response_with_external_editor(pending: PendingResponseEdit) -> tuple[bool, str]:
    editor = _resolve_editor_command()
    if editor is None:
        pending.keep_original()
        return False, "No editor found. Set $EDITOR (or install nano/vim/vi)."

    try:
        with tempfile.TemporaryDirectory(prefix="tproxy-edit-") as tmp_dir:
            tmp_path = Path(tmp_dir)
            header_file = tmp_path / "headers.txt"
            body_file = tmp_path / "body.txt"

            header_lines = [f"{name}: {value}" for name, value in pending.response.headers.items()]
            header_file.write_text("\n".join(header_lines) + "\n", encoding="utf-8")

            edit_plan = _build_body_edit_plan(
                headers=pending.response.headers,
                body=pending.response.body,
            )
            _write_body_file_for_editor(body_file=body_file, body_plan=edit_plan, original_body=pending.response.body)

            print(f"[tproxy] Editing response for {pending.request_url}")
            print(f"[tproxy] Opened {header_file.name} then {body_file.name} in {editor}")
            if edit_plan.is_text:
                print(f"[tproxy] Body edit mode: text (original charset: {edit_plan.charset})")
            else:
                print("[tproxy] Body edit mode: binary")

            _run_editor(editor, header_file)
            _run_editor(editor, body_file)

            edited_headers, edited_body = _read_edited_payload(
                header_file=header_file,
                body_file=body_file,
                body_plan=edit_plan,
            )

            pending.apply(headers=edited_headers, body=edited_body)
            saved_policy_name = _maybe_save_static_response_rule(
                pending=pending,
                headers=edited_headers,
                body=edited_body,
            )
            if saved_policy_name is not None:
                return True, f"Applied response edits and saved static policy {saved_policy_name}."
            return True, f"Applied response edits for {pending.request_url}"
    except Exception as exc:  # noqa: BLE001
        pending.keep_original()
        return False, f"Response edit failed ({exc}); kept original response."


def _build_body_edit_plan(*, headers: dict[str, str], body: bytes) -> _BodyEditPlan:
    body_is_text, body_charset = _is_textual_response_body(headers=headers, body=body)
    content_encoding = _header_value_case_insensitive(headers, "Content-Encoding").lower()
    if body_is_text and content_encoding and content_encoding != "identity":
        decoded = _decode_content_encoded_body(body, content_encoding)
        if decoded is not None:
            return _BodyEditPlan(
                is_text=True,
                charset=body_charset,
                editor_body=decoded,
                remove_content_encoding_header=True,
            )
        return _BodyEditPlan(
            is_text=False,
            charset=body_charset,
            editor_body=body,
            remove_content_encoding_header=False,
        )
    return _BodyEditPlan(
        is_text=body_is_text,
        charset=body_charset,
        editor_body=body,
        remove_content_encoding_header=False,
    )


def _write_body_file_for_editor(*, body_file: Path, body_plan: _BodyEditPlan, original_body: bytes) -> None:
    if body_plan.is_text:
        decoded = body_plan.editor_body.decode(body_plan.charset, errors="replace")
        body_file.write_text(decoded, encoding="utf-8")
        return
    body_file.write_bytes(original_body)


def _read_edited_payload(
    *,
    header_file: Path,
    body_file: Path,
    body_plan: _BodyEditPlan,
) -> tuple[dict[str, str], bytes]:
    edited_headers = _parse_header_file(header_file)
    if body_plan.is_text:
        edited_text = body_file.read_text(encoding="utf-8", errors="replace")
        edited_body = edited_text.encode(body_plan.charset, errors="replace")
        if body_plan.remove_content_encoding_header:
            _remove_header_case_insensitive(edited_headers, "Content-Encoding")
        return edited_headers, edited_body
    return edited_headers, body_file.read_bytes()


def _run_editor(editor: str, file_path: Path) -> None:
    cmd = shlex.split(editor) + [str(file_path)]
    subprocess.run(cmd, check=True)


def _parse_header_file(path: Path) -> dict[str, str]:
    headers: dict[str, str] = {}
    for raw_line in path.read_text(encoding="utf-8", errors="replace").splitlines():
        line = raw_line.strip()
        if not line:
            continue
        if line.startswith("#"):
            continue
        if ":" not in line:
            continue
        name, value = line.split(":", 1)
        headers[name.strip()] = value.strip()
    return headers


def _resolve_editor_command() -> str | None:
    env_editor = os.environ.get("EDITOR")
    if env_editor:
        parts = shlex.split(env_editor)
        if parts and shutil.which(parts[0]) is not None:
            return env_editor

    for candidate in ("nano", "vim", "vi"):
        if shutil.which(candidate) is not None:
            return candidate
    return None


def _is_textual_response_body(*, headers: dict[str, str], body: bytes) -> tuple[bool, str]:
    if not body:
        return True, "utf-8"

    content_type = _header_value_case_insensitive(headers, "Content-Type").lower()
    charset = _extract_charset(content_type) or "utf-8"

    textual_markers = (
        "text/",
        "application/json",
        "application/javascript",
        "application/x-javascript",
        "application/xml",
        "application/xhtml+xml",
        "application/x-www-form-urlencoded",
        "image/svg+xml",
    )
    if any(marker in content_type for marker in textual_markers):
        return True, charset

    if b"\x00" not in body:
        try:
            body.decode("utf-8")
            return True, "utf-8"
        except UnicodeDecodeError:
            pass
    return False, charset


def _extract_charset(content_type: str) -> str | None:
    for part in content_type.split(";")[1:]:
        token = part.strip()
        if not token.lower().startswith("charset="):
            continue
        value = token.split("=", 1)[1].strip().strip('"').strip("'")
        return value or None
    return None


def _header_value_case_insensitive(headers: dict[str, str], header_name: str) -> str:
    target = header_name.lower()
    for key, value in headers.items():
        if key.lower() == target:
            return value
    return ""


def _decode_content_encoded_body(body: bytes, content_encoding: str) -> bytes | None:
    encoding = content_encoding.strip().lower()
    if "," in encoding:
        encoding = encoding.split(",")[-1].strip()

    if encoding in {"", "identity"}:
        return body
    if encoding == "gzip":
        try:
            return gzip.decompress(body)
        except OSError:
            return None
    if encoding == "deflate":
        try:
            return zlib.decompress(body)
        except zlib.error:
            return None
    if encoding == "br":
        try:
            import brotli  # type: ignore[import-not-found]
        except Exception:
            return None
        try:
            return brotli.decompress(body)
        except Exception:
            return None
    return None


def _remove_header_case_insensitive(headers: dict[str, str], header_name: str) -> None:
    target = header_name.lower()
    for key in list(headers.keys()):
        if key.lower() == target:
            del headers[key]


def _maybe_save_static_response_rule(
    *,
    pending: PendingResponseEdit,
    headers: dict[str, str],
    body: bytes,
) -> str | None:
    try:
        answer = input(
            f"[tproxy] Save edited response as static policy for "
            f"{pending.method} {pending.request_url}? [y/N]: "
        ).strip().lower()
    except EOFError:
        return None
    if answer not in {"y", "yes"}:
        return None

    static_headers = dict(headers)
    _remove_header_case_insensitive(static_headers, "Transfer-Encoding")
    static_headers["Content-Length"] = str(len(body))

    runtime_config = get_runtime_config()
    try:
        policy_name = runtime_config.add_static_response_rule(
            url=pending.request_url,
            status_code=pending.response.status_code,
            reason=pending.response.reason,
            headers=static_headers,
            body=body,
            method=pending.method,
            url_prefix=False,
        )
        runtime_config.remove_open_editor_policy(pending.request_url, method=pending.method)
        return policy_name
    except ValueError:
        return None



from dataclasses import dataclass
import re
import selectors
import socket


_AUTHORITY_PATTERN = re.compile(r"^(?P<host>[A-Za-z0-9.\-]+):(?P<port>\d{1,5})$")


@dataclass(frozen=True)
class ConnectTarget:
    host: str
    port: int


class ConnectTunnelError(Exception):
    """Base error type for CONNECT tunnel setup/relay failures."""


class ConnectUpstreamConnectionError(ConnectTunnelError):
    """Raised when upstream TCP connection cannot be established."""


class ConnectUpstreamTimeoutError(ConnectTunnelError):
    """Raised when upstream connection or relay times out."""


def parse_connect_target(authority: str) -> ConnectTarget:
    """
    Parse CONNECT authority-form target: "<host>:<port>".
    """
    match = _AUTHORITY_PATTERN.fullmatch(authority)
    if not match:
        raise ValueError("Invalid CONNECT target. Expected '<host>:<port>'.")

    host = match.group("host")
    port = int(match.group("port"))
    if port < 1 or port > 65535:
        raise ValueError("Invalid CONNECT port. Must be between 1 and 65535.")

    return ConnectTarget(host=host, port=port)


def handle_connect_tunnel(
    *,
    client_socket: socket.socket,
    target: ConnectTarget,
    timeout_s: float = 30.0,
) -> None:
    """
    Handle HTTPS CONNECT tunneling.

    Socket roles:
    - client_socket: existing TCP connection (client <-> proxy)
    - upstream_socket: new TCP connection (proxy <-> target)

    Flow:
    1) Open upstream socket to requested target.
    2) Send 200 Connection Established back to client.
    3) Relay bytes bidirectionally until one side closes.
    """
    try:
        # Step 1: open a separate proxy->upstream TCP connection.
        upstream_socket = socket.create_connection((target.host, target.port), timeout=timeout_s)
    except socket.timeout as exc:
        raise ConnectUpstreamTimeoutError(
            f"Timed out connecting to upstream target {target.host}:{target.port}."
        ) from exc
    except OSError as exc:
        raise ConnectUpstreamConnectionError(
            f"Failed to connect to upstream target {target.host}:{target.port}: {exc}"
        ) from exc

    # Upstream connection succeeded; now upgrade the client connection into a tunnel.
    with upstream_socket:
        client_socket.sendall(b"HTTP/1.1 200 Connection Established\r\n\r\n")
        _relay_bidirectional(client_socket, upstream_socket, timeout_s=timeout_s)


def _relay_bidirectional(
    client_socket: socket.socket,
    upstream_socket: socket.socket,
    *,
    timeout_s: float,
) -> None:
    """
    Relay raw bytes client<->upstream until EOF or inactivity timeout.
    """
    client_socket.setblocking(False)
    upstream_socket.setblocking(False)

    selector = selectors.DefaultSelector()
    try:
        # For each readable socket, forward bytes to its opposite side.
        selector.register(client_socket, selectors.EVENT_READ, upstream_socket)
        selector.register(upstream_socket, selectors.EVENT_READ, client_socket)

        while True:
            events = selector.select(timeout=timeout_s)
            if not events:
                # Idle tunnel timeout: close gracefully instead of raising an HTTP-layer error.
                return

            for key, _mask in events:
                source_sock: socket.socket = key.fileobj
                destination_sock: socket.socket = key.data

                chunk = source_sock.recv(65536)

                # EOF on one side ends tunnel.
                if not chunk:
                    return

                destination_sock.sendall(chunk)
    finally:
        selector.close()



from dataclasses import dataclass
import logging
from queue import Empty, Queue
from threading import Event, Lock

from proxyscope.app.config.runtime import should_modify_response_for_request
from proxyscope.proxy.forwarding import ForwardResponse

LOGGER = logging.getLogger("tproxy.response_modifier")


@dataclass
class PendingResponseEdit:
    request_url: str
    method: str
    response: ForwardResponse
    _done: Event
    _result: ForwardResponse | None = None

    def apply(self, *, headers: dict[str, str], body: bytes) -> None:
        updated_headers = dict(headers)
        _remove_header_case_insensitive(updated_headers, "Transfer-Encoding")
        updated_headers["Content-Length"] = str(len(body))
        self._result = ForwardResponse(
            status_code=self.response.status_code,
            reason=self.response.reason,
            headers=updated_headers,
            body=body,
        )
        self._done.set()

    def keep_original(self) -> None:
        self._result = self.response
        self._done.set()

    def wait(self, *, timeout_s: float | None = None) -> ForwardResponse:
        finished = self._done.wait(timeout=timeout_s)
        if not finished or self._result is None:
            return self.response
        return self._result


class ResponseModifierService:
    def __init__(self, *, interactive_enabled: bool = False) -> None:
        self._interactive_enabled = interactive_enabled
        self._queue: Queue[PendingResponseEdit] = Queue()
        self._lock = Lock()

    def set_interactive_enabled(self, enabled: bool) -> None:
        with self._lock:
            self._interactive_enabled = enabled

    def maybe_modify_response(self, *, request_url: str, method: str, response: ForwardResponse) -> ForwardResponse:
        if not should_modify_response_for_request(method=method, url=request_url):
            return response

        with self._lock:
            interactive_enabled = self._interactive_enabled
        if not interactive_enabled:
            LOGGER.warning("Response modification requested for %s but interactive editor is disabled.", request_url)
            return response

        task = PendingResponseEdit(
            request_url=request_url,
            method=method,
            response=response,
            _done=Event(),
        )
        self._queue.put(task)
        return task.wait()

    def poll_pending_edit(self) -> PendingResponseEdit | None:
        try:
            return self._queue.get_nowait()
        except Empty:
            return None


_response_modifier = ResponseModifierService()
_response_modifier_lock = Lock()


def set_response_modifier(service: ResponseModifierService) -> None:
    global _response_modifier
    with _response_modifier_lock:
        _response_modifier = service


def get_response_modifier() -> ResponseModifierService:
    with _response_modifier_lock:
        return _response_modifier


def _remove_header_case_insensitive(headers: dict[str, str], header_name: str) -> None:
    target = header_name.lower()
    for key in list(headers.keys()):
        if key.lower() == target:
            del headers[key]



from collections.abc import Iterable
import json
import logging
from pathlib import Path
from threading import RLock

from proxyscope.app.config.matching import (
    first_rule_method,
    normalize_http_method,
    normalize_modification_url,
    normalize_whitelist_entry,
    policy_description,
    policy_rule_matches_request,
    request_url_candidates,
    rule_matches_url,
    rule_method_display,
    rule_url_display,
)
from proxyscope.app.config.models import PolicyRule, RequestMatchRule, StaticResponseTemplate
from proxyscope.app.config.serialization import (
    parse_policy_rule as parse_policy_rule_payload,
    serialize_policy_rule as serialize_policy_rule_payload,
)


class RuntimeConfig:
    """
    Runtime configuration shared by logging, policy matching, and UI.
    """

    def __init__(
        self,
        *,
        log_level: int = logging.INFO,
        log_whitelist: Iterable[str] | None = None,
        cache_invalidation_enabled: bool = True,
        config_path: str | Path | None = None,
        policy_rules: Iterable[PolicyRule] | None = None,
    ) -> None:
        self._lock = RLock()
        self._log_level = log_level
        self._log_whitelist: set[str] = set()
        self._cache_invalidation_enabled = cache_invalidation_enabled
        self._config_path = Path(config_path) if config_path is not None else None
        self._policy_rules: list[PolicyRule] = list(policy_rules or [])

        if log_whitelist is not None:
            for entry in log_whitelist:
                self._log_whitelist.add(normalize_whitelist_entry(entry))

    @property
    def config_path(self) -> Path | None:
        with self._lock:
            return self._config_path

    @property
    def log_level(self) -> int:
        with self._lock:
            return self._log_level

    def log_level_name(self) -> str:
        with self._lock:
            return logging.getLevelName(self._log_level)

    def set_log_level(self, value: str | int) -> int:
        with self._lock:
            if isinstance(value, int):
                new_level = value
            else:
                normalized = value.strip().upper()
                if not normalized:
                    raise ValueError("Log level must not be empty.")
                new_level = getattr(logging, normalized, None)
                if not isinstance(new_level, int):
                    raise ValueError(f"Unsupported log level: {value}")
            self._log_level = new_level
        self._persist_if_configured()
        return new_level

    def whitelist_entries(self) -> tuple[str, ...]:
        with self._lock:
            return tuple(sorted(self._log_whitelist))

    def add_whitelist_entry(self, value: str) -> str:
        normalized_host = normalize_whitelist_entry(value)
        with self._lock:
            self._log_whitelist.add(normalized_host)
        self._persist_if_configured()
        return normalized_host

    def remove_whitelist_entry(self, value: str) -> bool:
        normalized_host = normalize_whitelist_entry(value)
        removed = False
        with self._lock:
            if normalized_host in self._log_whitelist:
                self._log_whitelist.remove(normalized_host)
                removed = True
        if removed:
            self._persist_if_configured()
        return removed

    def clear_whitelist(self) -> None:
        with self._lock:
            self._log_whitelist.clear()
        self._persist_if_configured()

    def should_log_for_host(self, host: str | None) -> bool:
        with self._lock:
            if not self._log_whitelist:
                return True
            if host is None:
                return False
            normalized_host = normalize_whitelist_entry(host)
            return normalized_host in self._log_whitelist

    @property
    def cache_invalidation_enabled(self) -> bool:
        with self._lock:
            return self._cache_invalidation_enabled

    def set_cache_invalidation_enabled(self, enabled: bool) -> bool:
        with self._lock:
            self._cache_invalidation_enabled = enabled
        self._persist_if_configured()
        return enabled

    def toggle_cache_invalidation(self) -> bool:
        with self._lock:
            self._cache_invalidation_enabled = not self._cache_invalidation_enabled
            enabled = self._cache_invalidation_enabled
        self._persist_if_configured()
        return enabled

    def policy_rules(self) -> tuple[PolicyRule, ...]:
        with self._lock:
            return tuple(self._policy_rules)

    def set_policy_rules(self, rules: Iterable[PolicyRule]) -> None:
        with self._lock:
            self._policy_rules = list(rules)
        self._persist_if_configured()

    def add_policy_rule(self, rule: PolicyRule) -> None:
        with self._lock:
            self._policy_rules.append(rule)
        self._persist_if_configured()

    def clear_policy_rules(self) -> None:
        with self._lock:
            self._policy_rules.clear()
        self._persist_if_configured()

    def policy_descriptions(self) -> tuple[str, ...]:
        with self._lock:
            return tuple(policy_description(rule) for rule in self._policy_rules)

    def remove_policy_rule(self, name: str) -> bool:
        normalized = name.strip()
        if not normalized:
            raise ValueError("Policy name must not be empty.")
        removed = False
        with self._lock:
            kept: list[PolicyRule] = []
            for rule in self._policy_rules:
                if rule.name == normalized:
                    removed = True
                    continue
                kept.append(rule)
            self._policy_rules = kept
        if removed:
            self._persist_if_configured()
        return removed

    def get_policy_rule(self, name: str) -> PolicyRule | None:
        normalized = name.strip()
        if not normalized:
            raise ValueError("Policy name must not be empty.")
        with self._lock:
            for rule in self._policy_rules:
                if rule.name == normalized:
                    return rule
        return None

    def replace_policy_rule(self, name: str, replacement: PolicyRule) -> bool:
        normalized = name.strip()
        if not normalized:
            raise ValueError("Policy name must not be empty.")
        replaced = False
        with self._lock:
            updated: list[PolicyRule] = []
            for rule in self._policy_rules:
                if not replaced and rule.name == normalized:
                    updated.append(replacement)
                    replaced = True
                    continue
                updated.append(rule)
            self._policy_rules = updated
        if replaced:
            self._persist_if_configured()
        return replaced

    def set_policy_rule_enabled(self, name: str, *, enabled: bool) -> bool:
        normalized = name.strip()
        if not normalized:
            raise ValueError("Policy name must not be empty.")
        changed = False
        with self._lock:
            updated: list[PolicyRule] = []
            for rule in self._policy_rules:
                if rule.name != normalized:
                    updated.append(rule)
                    continue
                updated.append(
                    PolicyRule(
                        name=rule.name,
                        enabled=enabled,
                        action=rule.action,
                        match=rule.match,
                        static_response=rule.static_response,
                    )
                )
                changed = True
            self._policy_rules = updated
        if changed:
            self._persist_if_configured()
        return changed

    def add_static_response_rule(
        self,
        *,
        url: str,
        status_code: int = 200,
        reason: str = "OK",
        headers: dict[str, str] | None = None,
        body: bytes = b"",
        method: str | None = None,
        url_prefix: bool = False,
        name: str | None = None,
    ) -> str:
        normalized_url = normalize_modification_url(url)
        normalized_method = normalize_http_method(method) if method is not None else None
        rule_name = name or f"static-response-{len(self._policy_rules) + 1}"
        match = RequestMatchRule(
            methods=(normalized_method,) if normalized_method is not None else None,
            url_prefix=normalized_url if url_prefix else None,
            url_exact=None if url_prefix else normalized_url,
        )
        self.add_policy_rule(
            PolicyRule(
                name=rule_name,
                enabled=True,
                action="static_response",
                match=match,
                static_response=StaticResponseTemplate(
                    status_code=status_code,
                    reason=reason,
                    headers=dict(headers or {}),
                    body=body,
                ),
            )
        )
        return rule_name

    def modification_whitelist_entries(self) -> tuple[str, ...]:
        with self._lock:
            entries = [rule for rule in self._policy_rules if rule.action == "open_editor" and rule.enabled]
            return tuple(f"{rule_method_display(rule)} {rule_url_display(rule)}" for rule in entries)

    def open_editor_policy_entries(self) -> tuple[str, ...]:
        return self.modification_whitelist_entries()

    def add_open_editor_policy(self, value: str, *, method: str = "GET") -> str:
        return self.add_modification_whitelist_entry(value, method=method)

    def remove_open_editor_policy(self, value: str, *, method: str | None = None) -> bool:
        return self.remove_modification_whitelist_entry(value, method=method)

    def clear_open_editor_policies(self) -> None:
        self.clear_modification_whitelist()

    def add_modification_whitelist_entry(self, value: str, *, method: str = "GET") -> str:
        normalized_url = normalize_modification_url(value)
        normalized_method = normalize_http_method(method)
        name = f"open-editor-{len(self._policy_rules) + 1}"
        self.add_policy_rule(
            PolicyRule(
                name=name,
                enabled=True,
                action="open_editor",
                match=RequestMatchRule(methods=(normalized_method,), url_exact=normalized_url),
            )
        )
        return f"{normalized_method} {normalized_url}"

    def remove_modification_whitelist_entry(self, value: str, *, method: str | None = None) -> bool:
        normalized_url = normalize_modification_url(value)
        normalized_method = normalize_http_method(method) if method is not None else None
        removed = False
        with self._lock:
            kept: list[PolicyRule] = []
            for rule in self._policy_rules:
                if rule.action != "open_editor":
                    kept.append(rule)
                    continue
                if not rule_matches_url(rule, normalized_url):
                    kept.append(rule)
                    continue
                rule_method = first_rule_method(rule)
                if normalized_method is not None and rule_method != normalized_method:
                    kept.append(rule)
                    continue
                removed = True
            self._policy_rules = kept
        if removed:
            self._persist_if_configured()
        return removed

    def clear_modification_whitelist(self) -> None:
        with self._lock:
            self._policy_rules = [rule for rule in self._policy_rules if rule.action != "open_editor"]
        self._persist_if_configured()

    def should_modify_response_for_request(self, *, method: str, url: str) -> bool:
        normalized_method = normalize_http_method(method)
        candidates = request_url_candidates(url)
        with self._lock:
            for rule in self._policy_rules:
                if not rule.enabled or rule.action != "open_editor":
                    continue
                if policy_rule_matches_request(rule=rule, method=normalized_method, url_candidates=candidates):
                    return True
            return False

    def get_static_response_template_for_request(
        self,
        *,
        method: str,
        url: str,
    ) -> StaticResponseTemplate | None:
        normalized_method = normalize_http_method(method)
        candidates = request_url_candidates(url)
        with self._lock:
            for rule in self._policy_rules:
                if not rule.enabled or rule.action != "static_response":
                    continue
                if policy_rule_matches_request(rule=rule, method=normalized_method, url_candidates=candidates):
                    return rule.static_response
            return None

    def attach_config_path(self, path: str | Path) -> None:
        with self._lock:
            self._config_path = Path(path)

    def save_to_path(self, path: str | Path) -> Path:
        resolved = Path(path)
        with self._lock:
            self._config_path = resolved
        self.save()
        return resolved

    def reload_from_attached_file(self) -> bool:
        with self._lock:
            path = self._config_path
        if path is None:
            return False
        reloaded = RuntimeConfig.load_from_file(path)
        with self._lock:
            self._log_level = reloaded.log_level
            self._log_whitelist = set(reloaded.whitelist_entries())
            self._cache_invalidation_enabled = reloaded.cache_invalidation_enabled
            self._policy_rules = list(reloaded.policy_rules())
        return True

    def save(self) -> None:
        with self._lock:
            path = self._config_path
            payload = self._to_dict_locked()
        if path is None:
            return
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json.dumps(payload, indent=2, ensure_ascii=True) + "\n", encoding="utf-8")

    def _persist_if_configured(self) -> None:
        with self._lock:
            has_path = self._config_path is not None
        if has_path:
            self.save()

    def _to_dict_locked(self) -> dict:
        return {
            "log_level": logging.getLevelName(self._log_level),
            "log_whitelist": sorted(self._log_whitelist),
            "cache_invalidation_enabled": self._cache_invalidation_enabled,
            "policies": [serialize_policy_rule_payload(rule) for rule in self._policy_rules],
        }

    @classmethod
    def load_from_file(cls, path: str | Path) -> "RuntimeConfig":
        resolved_path = Path(path)
        if not resolved_path.exists():
            return cls(config_path=resolved_path)

        raw = json.loads(resolved_path.read_text(encoding="utf-8"))
        level_raw = str(raw.get("log_level", "INFO")).strip().upper()
        level_value = getattr(logging, level_raw, logging.INFO)
        whitelist_raw = raw.get("log_whitelist", [])
        cache_enabled = bool(raw.get("cache_invalidation_enabled", False))

        rules: list[PolicyRule] = []
        for item in raw.get("policies", []):
            parsed = parse_policy_rule_payload(item)
            if parsed is not None:
                rules.append(parsed)

        return cls(
            log_level=level_value,
            log_whitelist=whitelist_raw,
            cache_invalidation_enabled=cache_enabled,
            config_path=resolved_path,
            policy_rules=rules,
        )


_runtime_config = RuntimeConfig()
_runtime_config_lock = RLock()


def set_runtime_config(config: RuntimeConfig) -> None:
    global _runtime_config
    with _runtime_config_lock:
        _runtime_config = config


def get_runtime_config() -> RuntimeConfig:
    with _runtime_config_lock:
        return _runtime_config


def should_log_for_host(host: str | None) -> bool:
    return get_runtime_config().should_log_for_host(host)


def is_cache_invalidation_enabled() -> bool:
    return get_runtime_config().cache_invalidation_enabled


def should_modify_response_for_request(*, method: str, url: str) -> bool:
    return get_runtime_config().should_modify_response_for_request(method=method, url=url)


def get_static_response_template_for_request(*, method: str, url: str) -> StaticResponseTemplate | None:
    return get_runtime_config().get_static_response_template_for_request(method=method, url=url)


def serialize_policy_rule(rule: PolicyRule) -> dict:
    return serialize_policy_rule_payload(rule)


def parse_policy_rule(data: object) -> PolicyRule | None:
    return parse_policy_rule_payload(data)



import logging
import time
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Final
from urllib.parse import urlsplit

from proxyscope.proxy.connect_tunnel import (
    ConnectUpstreamConnectionError,
    ConnectUpstreamTimeoutError,
    handle_connect_tunnel,
    parse_connect_target,
)
from proxyscope.proxy.forwarding import (
    ForwardRequest,
    ForwardResponse,
    UpstreamForwarder,
    prepare_forward_headers,
    prepare_forward_request,
    resolve_target_url,
)
from proxyscope.proxy.http_bridge import map_incoming_request, write_forward_response
from proxyscope.mitm.certificates import MitmCertificateError, default_ca
from proxyscope.mitm.tunnel import MitmTLSInterceptor
from proxyscope.app.logging.observability import emit_site_visit
from proxyscope.app.config.runtime import get_static_response_template_for_request
from proxyscope.app.editing.modifier import get_response_modifier
from proxyscope.proxy.types import Forwarder
from proxyscope.proxy.tunnel_registry import TunnelConnectionRegistry
from proxyscope.app.logging.request_response import REQUEST_LOGGER, log_incoming_request, log_outgoing_response

SERVER_LOGGER: Final = logging.getLogger("tproxy.server")


class ProxyHTTPServer(ThreadingHTTPServer):
    def __init__(
        self,
        server_address: tuple[str, int],
        request_handler_class: type[BaseHTTPRequestHandler],
        *,
        forwarder: Forwarder,
        mitm_interceptor: MitmTLSInterceptor | None = None,
    ) -> None:
        super().__init__(server_address, request_handler_class)
        self.forwarder = forwarder
        self.mitm_interceptor = mitm_interceptor
        self._tunnel_registry = TunnelConnectionRegistry()

    def register_tunnel_socket(self, sock: object) -> None:
        self._tunnel_registry.register(sock)

    def unregister_tunnel_socket(self, sock: object) -> None:
        self._tunnel_registry.unregister(sock)

    def close_all_active_tunnels(self) -> int:
        return self._tunnel_registry.close_all()


class RequestLoggingHandler(BaseHTTPRequestHandler):
    server_version = "tproxy/0.1"

    def _send_connect_error_response(self, *, status: int, reason: str, body: bytes) -> None:
        """
        Best-effort HTTP error response for CONNECT setup failures.
        """
        try:
            self.send_response(status, reason)
            self.send_header("Content-Type", "text/plain; charset=utf-8")
            self.send_header("Content-Length", str(len(body)))
            self.end_headers()
            self.wfile.write(body)
        except OSError:
            # Connection may already be closed or switched out of HTTP framing.
            SERVER_LOGGER.debug("Could not write CONNECT error response; socket no longer writable.")

    def _log_connect_outcome(
        self,
        *,
        request_id: int | None,
        started: float,
        target_host: str | None,
        status_code: int,
        reason: str,
        body: bytes = b"",
        headers: dict[str, str] | None = None,
    ) -> None:
        duration_ms = (time.perf_counter() - started) * 1000
        response_headers = dict(headers or {})
        log_outgoing_response(
            ForwardResponse(
                status_code=status_code,
                reason=reason,
                headers=response_headers,
                body=body,
            ),
            request_id=request_id,
            duration_ms=duration_ms,
            client_ip=self.client_address[0],
            target_host=target_host,
        )

    def _respond_connect_error_and_log(
        self,
        *,
        request_id: int | None,
        started: float,
        target_host: str | None,
        status: int,
        reason: str,
        body: bytes,
    ) -> None:
        self._send_connect_error_response(status=status, reason=reason, body=body)
        self._log_connect_outcome(
            request_id=request_id,
            started=started,
            target_host=target_host,
            status_code=status,
            reason=reason,
            body=body,
            headers={
                "Content-Type": "text/plain; charset=utf-8",
                "Content-Length": str(len(body)),
            },
        )

    def _handle(self, *, send_body: bool) -> None:
        started = time.perf_counter()
        server = self.server
        if not isinstance(server, ProxyHTTPServer):
            self.send_error(500, "Server misconfiguration")
            return

        forward_request = map_incoming_request(self)
        forward_request = prepare_forward_request(forward_request)
        effective_headers = prepare_forward_headers(forward_request.headers)
        forward_request = ForwardRequest(
            method=forward_request.method,
            path=forward_request.path,
            headers=effective_headers,
            body=forward_request.body,
        )
        target_host, target_port = _resolve_forward_request_target(forward_request)
        request_id = log_incoming_request(
            method=self.command,
            path=forward_request.path,
            client_ip=self.client_address[0],
            headers=effective_headers,
            body=forward_request.body,
            target_host=target_host,
            target_port=target_port,
        )
        try:
            if target_host:
                emit_site_visit(target_host)
        except ValueError:
            pass
        try:
            target_url = resolve_target_url(forward_request)
        except ValueError as exc:
            error_body = f"{exc}\n".encode("utf-8")
            self.send_response(400, "Bad Request")
            self.send_header("Content-Type", "text/plain; charset=utf-8")
            self.send_header("Content-Length", str(len(error_body)))
            self.end_headers()
            if send_body:
                self.wfile.write(error_body)
            duration_ms = (time.perf_counter() - started) * 1000
            log_outgoing_response(
                ForwardResponse(
                    status_code=400,
                    reason="Bad Request",
                    headers={
                        "Content-Type": "text/plain; charset=utf-8",
                        "Content-Length": str(len(error_body)),
                    },
                    body=error_body,
                ),
                request_id=request_id,
                duration_ms=duration_ms,
                client_ip=self.client_address[0],
                target_host=target_host,
            )
            return

        static_template = get_static_response_template_for_request(method=self.command, url=target_url)
        if static_template is not None:
            forward_response = ForwardResponse(
                status_code=static_template.status_code,
                reason=static_template.reason,
                headers=dict(static_template.headers),
                body=static_template.body,
            )
        else:
            forward_response = server.forwarder.forward(forward_request)
        modifier = get_response_modifier()
        forward_response = modifier.maybe_modify_response(
            request_url=target_url,
            method=self.command,
            response=forward_response,
        )
        write_forward_response(self, forward_response, send_body=send_body)

        duration_ms = (time.perf_counter() - started) * 1000
        log_outgoing_response(
            forward_response,
            request_id=request_id,
            duration_ms=duration_ms,
            client_ip=self.client_address[0],
            target_host=target_host,
        )

    def do_GET(self) -> None:
        self._handle(send_body=True)

    def do_POST(self) -> None:
        self._handle(send_body=True)

    def do_PUT(self) -> None:
        self._handle(send_body=True)

    def do_PATCH(self) -> None:
        self._handle(send_body=True)

    def do_DELETE(self) -> None:
        self._handle(send_body=True)

    def do_OPTIONS(self) -> None:
        self._handle(send_body=True)

    def do_HEAD(self) -> None:
        self._handle(send_body=False)

    def do_CONNECT(self) -> None:
        started = time.perf_counter()
        target_host: str | None = None
        target_port: int | None = None
        request_id: int | None = None
        try:
            target = parse_connect_target(self.path)
            target_host = target.host
            target_port = target.port
            request_id = log_incoming_request(
                method=self.command,
                path=self.path,
                client_ip=self.client_address[0],
                headers={name: value for name, value in self.headers.items()},
                body=b"",
                target_host=target_host,
                target_port=target_port,
            )
            emit_site_visit(target.host)
            server = self.server
            if not isinstance(server, ProxyHTTPServer):
                self.send_error(500, "Server misconfiguration")
                return
            server.register_tunnel_socket(self.connection)

            if server.mitm_interceptor is not None:
                tunnel_established = server.mitm_interceptor.intercept(
                    client_socket=self.connection,
                    target=target,
                    timeout_s=30.0,
                )
                if not tunnel_established:
                    REQUEST_LOGGER.warning(
                        "MITM client TLS handshake failed target=%s:%d client=%s",
                        target.host,
                        target.port,
                        self.client_address[0],
                    )
                    self._log_connect_outcome(
                        request_id=request_id,
                        started=started,
                        target_host=target_host,
                        status_code=495,
                        reason="Client TLS Handshake Failed",
                    )
                    return
            else:
                handle_connect_tunnel(
                    client_socket=self.connection,
                    target=target,
                    timeout_s=30.0,
                )
            self._log_connect_outcome(
                request_id=request_id,
                started=started,
                target_host=target_host,
                status_code=200,
                reason="Connection Established",
            )
            return
        except ValueError as exc:
            if request_id is None:
                request_id = log_incoming_request(
                    method=self.command,
                    path=self.path,
                    client_ip=self.client_address[0],
                    headers={name: value for name, value in self.headers.items()},
                    body=b"",
                    target_host=target_host,
                    target_port=target_port,
                )
            response_body = f"Invalid CONNECT target: {exc}\n".encode("utf-8")
            self._respond_connect_error_and_log(
                request_id=request_id,
                started=started,
                target_host=target_host,
                status=400,
                reason="Bad CONNECT target",
                body=response_body,
            )
            return
        except ConnectUpstreamTimeoutError:
            response_body = b"CONNECT upstream timed out.\n"
            self._respond_connect_error_and_log(
                request_id=request_id,
                started=started,
                target_host=target_host,
                status=504,
                reason="Gateway Timeout",
                body=response_body,
            )
            return
        except ConnectUpstreamConnectionError:
            response_body = b"CONNECT upstream connection failed.\n"
            self._respond_connect_error_and_log(
                request_id=request_id,
                started=started,
                target_host=target_host,
                status=502,
                reason="Bad Gateway",
                body=response_body,
            )
            return
        finally:
            server = self.server
            if isinstance(server, ProxyHTTPServer):
                server.unregister_tunnel_socket(self.connection)

    def log_message(self, fmt: str, *args: object) -> None:
        # Route default request logs through the project logger.
        SERVER_LOGGER.debug(fmt, *args)


def create_server(
    host: str,
    port: int,
    *,
    forwarder: Forwarder | None = None,
    mitm_interceptor: MitmTLSInterceptor | None = None,
    auto_enable_mitm: bool = True,
) -> ThreadingHTTPServer:
    resolved_forwarder = forwarder or UpstreamForwarder()
    resolved_mitm_interceptor = mitm_interceptor
    if resolved_mitm_interceptor is None and auto_enable_mitm:
        ca = default_ca()
        try:
            created_new_ca = ca.ensure_ca_material()
            if created_new_ca:
                SERVER_LOGGER.info(
                    "Generated local MITM CA materials cert=%s key=%s",
                    ca.ca_cert_path,
                    ca.ca_key_path,
                )
            resolved_mitm_interceptor = MitmTLSInterceptor(certificate_authority=ca)
        except MitmCertificateError as exc:
            SERVER_LOGGER.warning("MITM disabled: failed to initialize local CA: %s", exc)

    return ProxyHTTPServer(
        (host, port),
        RequestLoggingHandler,
        forwarder=resolved_forwarder,
        mitm_interceptor=resolved_mitm_interceptor,
    )


def _resolve_forward_request_target(forward_request: ForwardRequest) -> tuple[str | None, int | None]:
    try:
        target_url = resolve_target_url(forward_request)
    except ValueError:
        return None, None

    parsed = urlsplit(target_url)
    host = parsed.hostname
    if host is None:
        return None, None

    if parsed.port is not None:
        return host, parsed.port
    if parsed.scheme == "https":
        return host, 443
    return host, 80



from dataclasses import dataclass


@dataclass(frozen=True)
class RequestMatchRule:
    methods: tuple[str, ...] | None = None
    url_exact: str | None = None
    url_prefix: str | None = None


@dataclass(frozen=True)
class StaticResponseTemplate:
    status_code: int
    reason: str
    headers: dict[str, str]
    body: bytes


@dataclass(frozen=True)
class PolicyRule:
    name: str
    enabled: bool
    action: str  # "open_editor" | "static_response"
    match: RequestMatchRule
    static_response: StaticResponseTemplate | None = None



from dataclasses import dataclass, replace
from threading import RLock
from time import time


@dataclass(frozen=True)
class LoggedRequestMessage:
    method: str
    path: str
    start_line: str
    headers: tuple[tuple[str, str], ...]
    body_preview: str


@dataclass(frozen=True)
class LoggedResponseMessage:
    status_code: int
    reason: str
    start_line: str
    headers: tuple[tuple[str, str], ...]
    body_preview: str


@dataclass(frozen=True)
class LoggedExchange:
    request_id: int
    started_at: float
    finished_at: float | None
    duration_ms: float | None
    client_ip: str
    target_host: str | None
    target_port: int | None
    protocol: str
    request: LoggedRequestMessage
    response: LoggedResponseMessage | None


class RequestJournal:
    """
    Central in-memory request/response log storage for the runtime UI.
    """

    def __init__(self, *, max_entries: int = 1000) -> None:
        self._lock = RLock()
        self._max_entries = max_entries
        self._next_id = 1
        self._entries: list[LoggedExchange] = []

    def start_request(
        self,
        *,
        method: str,
        path: str,
        start_line: str,
        headers: dict[str, str],
        body: bytes | None,
        client_ip: str,
        target_host: str | None,
        target_port: int | None,
        protocol: str,
    ) -> int:
        with self._lock:
            request_id = self._next_id
            self._next_id += 1
            exchange = LoggedExchange(
                request_id=request_id,
                started_at=time(),
                finished_at=None,
                duration_ms=None,
                client_ip=client_ip,
                target_host=target_host,
                target_port=target_port,
                protocol=protocol,
                request=LoggedRequestMessage(
                    method=method,
                    path=path,
                    start_line=start_line,
                    headers=tuple(headers.items()),
                    body_preview=_body_preview(body),
                ),
                response=None,
            )
            self._entries.append(exchange)
            if len(self._entries) > self._max_entries:
                self._entries = self._entries[-self._max_entries :]
            return request_id

    def complete_request(
        self,
        request_id: int,
        *,
        status_code: int,
        reason: str,
        start_line: str,
        headers: dict[str, str],
        body: bytes | None,
        duration_ms: float,
    ) -> None:
        with self._lock:
            for index, entry in enumerate(self._entries):
                if entry.request_id != request_id:
                    continue
                updated = replace(
                    entry,
                    finished_at=time(),
                    duration_ms=duration_ms,
                    response=LoggedResponseMessage(
                        status_code=status_code,
                        reason=reason,
                        start_line=start_line,
                        headers=tuple(headers.items()),
                        body_preview=_body_preview(body),
                    ),
                )
                self._entries[index] = updated
                return

    def list_entries(self) -> tuple[LoggedExchange, ...]:
        with self._lock:
            return tuple(self._entries)

    def get_entry(self, request_id: int) -> LoggedExchange | None:
        with self._lock:
            for entry in self._entries:
                if entry.request_id == request_id:
                    return entry
        return None

    def clear(self) -> None:
        with self._lock:
            self._entries.clear()


_request_journal = RequestJournal()
_request_journal_lock = RLock()


def set_request_journal(journal: RequestJournal) -> None:
    global _request_journal
    with _request_journal_lock:
        _request_journal = journal


def get_request_journal() -> RequestJournal:
    with _request_journal_lock:
        return _request_journal


def _body_preview(body: bytes | None, *, max_bytes: int = 4096) -> str:
    if body is None:
        return "<not captured>"
    if not body:
        return ""

    prefix = body[:max_bytes]
    text = prefix.decode("utf-8", errors="replace").replace("\x00", "\\x00")
    if len(body) <= max_bytes:
        return text
    remaining = len(body) - max_bytes
    return f"{text}\n...[truncated {remaining} bytes]"



from threading import Lock
from typing import Protocol


class RuntimeObserver(Protocol):
    def on_site_visit(self, host: str) -> None:
        ...


_observer_lock = Lock()
_observer: RuntimeObserver | None = None


def set_runtime_observer(observer: RuntimeObserver | None) -> None:
    global _observer
    with _observer_lock:
        _observer = observer


def emit_site_visit(host: str) -> None:
    with _observer_lock:
        observer = _observer
    if observer is None:
        return
    observer.on_site_visit(host)




import logging
from typing import Final

from proxyscope.app.config.runtime import should_log_for_host
from proxyscope.app.runtime.journal import get_request_journal
from proxyscope.proxy.forwarding import ForwardResponse

REQUEST_LOGGER: Final = logging.getLogger("tproxy.request")
RESPONSE_LOGGER: Final = logging.getLogger("tproxy.response")


def log_incoming_request(
    *,
    method: str,
    path: str,
    client_ip: str,
    headers: dict[str, str],
    body: bytes = b"",
    target_host: str | None = None,
    target_port: int | None = None,
    protocol: str = "http",
) -> int | None:
    if not should_log_for_host(target_host):
        return None

    REQUEST_LOGGER.info(
        "Incoming request method=%s path=%s client=%s target=%s",
        method,
        path,
        client_ip,
        _target_display(target_host, target_port),
    )

    journal = get_request_journal()
    return journal.start_request(
        method=method,
        path=path,
        start_line=f"{method} {path}",
        headers=headers,
        body=body,
        client_ip=client_ip,
        target_host=target_host,
        target_port=target_port,
        protocol=protocol,
    )


def log_outgoing_response(
    response: ForwardResponse,
    *,
    request_id: int | None,
    duration_ms: float,
    client_ip: str,
    target_host: str | None = None,
) -> None:
    if not should_log_for_host(target_host):
        return

    RESPONSE_LOGGER.info(
        "Outgoing response status=%s reason=%s client=%s body_bytes=%d duration_ms=%.2f",
        response.status_code,
        response.reason,
        client_ip,
        len(response.body),
        duration_ms,
    )

    if request_id is None:
        return

    journal = get_request_journal()
    journal.complete_request(
        request_id,
        status_code=response.status_code,
        reason=response.reason,
        start_line=f"HTTP/1.1 {response.status_code} {response.reason}",
        headers=response.headers,
        body=response.body,
        duration_ms=duration_ms,
    )


def log_mitm_http_request(
    *,
    client_ip: str,
    target_host: str,
    target_port: int,
    start_line: str,
    headers: dict[str, str],
    body: bytes,
) -> int | None:
    if not should_log_for_host(target_host):
        return None

    REQUEST_LOGGER.info(
        "MITM HTTP request client=%s target=%s:%d start_line=%s",
        client_ip,
        target_host,
        target_port,
        start_line,
    )

    method, path = _parse_request_start_line(start_line)
    journal = get_request_journal()
    return journal.start_request(
        method=method,
        path=path,
        start_line=start_line,
        headers=headers,
        body=body,
        client_ip=client_ip,
        target_host=target_host,
        target_port=target_port,
        protocol="https-mitm",
    )


def log_mitm_http_response(
    *,
    client_ip: str,
    target_host: str,
    target_port: int,
    start_line: str,
    headers: dict[str, str],
    body: bytes,
    request_id: int | None,
) -> None:
    if not should_log_for_host(target_host):
        return

    RESPONSE_LOGGER.info(
        "MITM HTTP response client=%s target=%s:%d start_line=%s",
        client_ip,
        target_host,
        target_port,
        start_line,
    )

    if request_id is None:
        return

    status_code, reason = _parse_response_start_line(start_line)
    journal = get_request_journal()
    journal.complete_request(
        request_id,
        status_code=status_code,
        reason=reason,
        start_line=start_line,
        headers=headers,
        body=body,
        duration_ms=0.0,
    )


def _parse_request_start_line(start_line: str) -> tuple[str, str]:
    parts = start_line.split(" ", 2)
    if len(parts) >= 2:
        return parts[0], parts[1]
    return "UNKNOWN", start_line


def _parse_response_start_line(start_line: str) -> tuple[int, str]:
    parts = start_line.split(" ", 2)
    if len(parts) >= 2:
        try:
            status_code = int(parts[1])
        except ValueError:
            status_code = 0
        reason = parts[2] if len(parts) >= 3 else ""
        return status_code, reason
    return 0, ""


def _target_display(host: str | None, port: int | None) -> str:
    if host is None:
        return "-"
    if port is None:
        return host
    return f"{host}:{port}"



from collections.abc import Callable


class HTTP1MessageSniffer:
    """
    Incrementally parse HTTP/1.x messages and emit start-line + headers + body.
    Body capture is capped to avoid unbounded memory growth.
    """

    def __init__(
        self,
        on_message: Callable[[str, dict[str, str], bytes], None],
        *,
        max_body_capture_bytes: int = 8192,
    ) -> None:
        self._on_message = on_message
        self._max_body_capture_bytes = max_body_capture_bytes
        self._buffer = bytearray()
        self._mode = "headers"
        self._body_remaining = 0
        self._current_start_line = ""
        self._current_headers: dict[str, str] = {}
        self._current_body = bytearray()

    def feed(self, data: bytes) -> None:
        self._buffer.extend(data)
        while True:
            if self._mode == "headers":
                header_end = self._find_header_end()
                if header_end < 0:
                    return

                start_line, headers = self._consume_headers(header_end)
                if not start_line:
                    continue

                self._current_start_line = start_line
                self._current_headers = headers
                self._current_body.clear()

                transfer_encoding = _header_value(headers, "transfer-encoding").lower()
                if "chunked" in transfer_encoding:
                    self._mode = "chunked"
                    continue

                content_length_value = _header_value(headers, "content-length")
                if content_length_value:
                    try:
                        self._body_remaining = max(0, int(content_length_value))
                    except ValueError:
                        self._body_remaining = 0
                else:
                    self._body_remaining = 0

                if self._body_remaining > 0:
                    self._mode = "content-length"
                    continue

                self._emit_current_message()
                self._mode = "headers"
                continue

            if self._mode == "content-length":
                consumed = self._consume_content_length_body()
                if not consumed:
                    return
                self._emit_current_message()
                self._mode = "headers"
                continue

            if self._mode == "chunked":
                if not self._consume_chunked_body():
                    return
                self._emit_current_message()
                self._mode = "headers"
                continue

            return

    def _emit_current_message(self) -> None:
        self._on_message(self._current_start_line, self._current_headers, bytes(self._current_body))

    def _find_header_end(self) -> int:
        return self._buffer.find(b"\r\n\r\n")

    def _consume_headers(self, header_end: int) -> tuple[str, dict[str, str]]:
        raw_header_block = bytes(self._buffer[:header_end])
        del self._buffer[: header_end + 4]

        lines = raw_header_block.decode("iso-8859-1", errors="replace").split("\r\n")
        if not lines:
            return "", {}

        start_line = lines[0].strip()
        headers: dict[str, str] = {}
        for line in lines[1:]:
            if not line or ":" not in line:
                continue
            name, value = line.split(":", 1)
            headers[name.strip()] = value.strip()
        return start_line, headers

    def _consume_content_length_body(self) -> bool:
        if len(self._buffer) < self._body_remaining:
            return False

        if self._body_remaining > 0 and len(self._current_body) < self._max_body_capture_bytes:
            to_copy = min(self._body_remaining, self._max_body_capture_bytes - len(self._current_body))
            self._current_body.extend(self._buffer[:to_copy])

        del self._buffer[: self._body_remaining]
        self._body_remaining = 0
        return True

    def _consume_chunked_body(self) -> bool:
        while True:
            line_end = self._buffer.find(b"\r\n")
            if line_end < 0:
                return False

            chunk_size_line = bytes(self._buffer[:line_end])
            try:
                chunk_size = int(chunk_size_line.split(b";", 1)[0], 16)
            except ValueError:
                self._buffer.clear()
                return False

            if chunk_size == 0:
                size_line_total = line_end + 2
                if len(self._buffer) < size_line_total:
                    return False
                del self._buffer[:size_line_total]

                if self._buffer.startswith(b"\r\n"):
                    del self._buffer[:2]
                    while self._buffer.startswith(b"\r\n"):
                        del self._buffer[:2]
                    return True

                trailer_end = self._buffer.find(b"\r\n\r\n")
                if trailer_end < 0:
                    return False
                del self._buffer[: trailer_end + 4]
                return True

            total_needed = line_end + 2 + chunk_size + 2
            if len(self._buffer) < total_needed:
                return False

            chunk_start = line_end + 2
            if chunk_size > 0 and len(self._current_body) < self._max_body_capture_bytes:
                remaining = self._max_body_capture_bytes - len(self._current_body)
                self._current_body.extend(self._buffer[chunk_start : chunk_start + min(chunk_size, remaining)])

            del self._buffer[:total_needed]

def _header_value(headers: dict[str, str], name: str) -> str:
    target = name.lower()
    for key, value in headers.items():
        if key.lower() == target:
            return value
    return ""



class HTTP1RequestHeaderRewriter:
    """
    Stream rewriter for HTTP/1.x requests.
    Rewrites request headers while preserving message framing and body bytes.
    """

    def __init__(self) -> None:
        self._buffer = bytearray()
        self._mode = "headers"
        self._body_remaining = 0

    def feed(self, data: bytes) -> bytes:
        self._buffer.extend(data)
        out = bytearray()

        while True:
            if self._mode == "headers":
                header_end = self._buffer.find(b"\r\n\r\n")
                if header_end < 0:
                    break

                raw_header_block = bytes(self._buffer[:header_end])
                del self._buffer[: header_end + 4]

                rewritten_headers_block, headers = _rewrite_header_block(raw_header_block)
                out.extend(rewritten_headers_block)

                transfer_encoding = _header_value(headers, "transfer-encoding").lower()
                if "chunked" in transfer_encoding:
                    self._mode = "chunked"
                    continue

                content_length = _header_value(headers, "content-length")
                if content_length:
                    try:
                        self._body_remaining = max(0, int(content_length))
                    except ValueError:
                        self._body_remaining = 0
                else:
                    self._body_remaining = 0

                if self._body_remaining > 0:
                    self._mode = "content-length"
                    continue

                self._mode = "headers"
                continue

            if self._mode == "content-length":
                if not self._buffer:
                    break
                to_copy = min(len(self._buffer), self._body_remaining)
                out.extend(self._buffer[:to_copy])
                del self._buffer[:to_copy]
                self._body_remaining -= to_copy
                if self._body_remaining <= 0:
                    self._mode = "headers"
                continue

            if self._mode == "chunked":
                progressed = self._consume_chunked(out)
                if not progressed:
                    break
                continue

            break

        return bytes(out)

    def flush(self) -> bytes:
        """
        Flush buffered bytes (for abrupt stream termination cases).
        """
        if not self._buffer:
            return b""
        data = bytes(self._buffer)
        self._buffer.clear()
        return data

    def _consume_chunked(self, out: bytearray) -> bool:
        line_end = self._buffer.find(b"\r\n")
        if line_end < 0:
            return False

        chunk_size_line = bytes(self._buffer[:line_end])
        try:
            chunk_size = int(chunk_size_line.split(b";", 1)[0], 16)
        except ValueError:
            # Malformed chunked stream, pass through what is buffered and reset.
            out.extend(self._buffer)
            self._buffer.clear()
            self._mode = "headers"
            return True

        if chunk_size == 0:
            trailer_end = self._buffer.find(b"\r\n\r\n", line_end + 2)
            if trailer_end < 0:
                return False
            segment_end = trailer_end + 4
            out.extend(self._buffer[:segment_end])
            del self._buffer[:segment_end]
            self._mode = "headers"
            return True

        total_needed = line_end + 2 + chunk_size + 2
        if len(self._buffer) < total_needed:
            return False

        out.extend(self._buffer[:total_needed])
        del self._buffer[:total_needed]
        return True


def rewrite_cache_invalidation_headers(headers: dict[str, str]) -> dict[str, str]:
    rewritten: dict[str, str] = {}
    blocked = {
        "if-none-match",
        "if-modified-since",
        "cache-control",
        "pragma",
        "expires",
        "accept-encoding",
    }
    for key, value in headers.items():
        if key.lower() in blocked:
            continue
        rewritten[key] = value

    rewritten["Cache-Control"] = "no-cache, no-store, max-age=0, must-revalidate"
    rewritten["Pragma"] = "no-cache"
    rewritten["Expires"] = "0"
    # Keep upstream payloads uncompressed so TUI preview/editor can work with text bodies.
    rewritten["Accept-Encoding"] = "identity"
    return rewritten


def _rewrite_header_block(raw_header_block: bytes) -> tuple[bytes, dict[str, str]]:
    lines = raw_header_block.decode("iso-8859-1", errors="replace").split("\r\n")
    if not lines:
        return raw_header_block + b"\r\n\r\n", {}

    start_line = lines[0]
    parsed_headers: dict[str, str] = {}
    ordered: list[tuple[str, str]] = []
    for line in lines[1:]:
        if not line or ":" not in line:
            continue
        name, value = line.split(":", 1)
        name_stripped = name.strip()
        value_stripped = value.strip()
        ordered.append((name_stripped, value_stripped))
        parsed_headers[name_stripped] = value_stripped

    rewritten_headers = rewrite_cache_invalidation_headers(dict(ordered))
    rebuilt_lines = [start_line]
    for name, value in rewritten_headers.items():
        rebuilt_lines.append(f"{name}: {value}")
    rebuilt = "\r\n".join(rebuilt_lines).encode("iso-8859-1", errors="replace") + b"\r\n\r\n"
    return rebuilt, rewritten_headers


def _header_value(headers: dict[str, str], name: str) -> str:
    target = name.lower()
    for key, value in headers.items():
        if key.lower() == target:
            return value
    return ""



import base64

from proxyscope.app.config.matching import normalize_http_method, normalize_modification_url
from proxyscope.app.config.models import PolicyRule, RequestMatchRule, StaticResponseTemplate


def serialize_policy_rule(rule: PolicyRule) -> dict:
    payload = {
        "name": rule.name,
        "enabled": rule.enabled,
        "action": {"type": rule.action},
        "match": {},
    }
    if rule.match.methods is not None:
        payload["match"]["methods"] = list(rule.match.methods)
    if rule.match.url_exact is not None:
        payload["match"]["url_exact"] = rule.match.url_exact
    if rule.match.url_prefix is not None:
        payload["match"]["url_prefix"] = rule.match.url_prefix

    if rule.action == "static_response" and rule.static_response is not None:
        payload["action"]["status_code"] = rule.static_response.status_code
        payload["action"]["reason"] = rule.static_response.reason
        payload["action"]["headers"] = dict(rule.static_response.headers)
        try:
            payload["action"]["body"] = rule.static_response.body.decode("utf-8")
        except UnicodeDecodeError:
            payload["action"]["body_base64"] = base64.b64encode(rule.static_response.body).decode("ascii")
    return payload


def parse_policy_rule(data: object) -> PolicyRule | None:
    if not isinstance(data, dict):
        return None
    name = str(data.get("name", "unnamed-policy"))
    enabled = bool(data.get("enabled", True))

    action_data = data.get("action", {})
    if not isinstance(action_data, dict):
        return None
    action_type = str(action_data.get("type", "")).strip()
    if action_type not in {"open_editor", "static_response"}:
        return None

    match_data = data.get("match", {})
    if not isinstance(match_data, dict):
        return None
    methods_raw = match_data.get("methods")
    methods: tuple[str, ...] | None = None
    if isinstance(methods_raw, list):
        normalized_methods = [normalize_http_method(str(item)) for item in methods_raw]
        methods = tuple(normalized_methods) if normalized_methods else None
    elif isinstance(methods_raw, str):
        methods = (normalize_http_method(methods_raw),)

    url_exact = match_data.get("url_exact")
    url_prefix = match_data.get("url_prefix")
    normalized_exact = normalize_modification_url(url_exact) if isinstance(url_exact, str) else None
    normalized_prefix = normalize_modification_url(url_prefix) if isinstance(url_prefix, str) else None

    match = RequestMatchRule(methods=methods, url_exact=normalized_exact, url_prefix=normalized_prefix)

    if action_type == "open_editor":
        return PolicyRule(name=name, enabled=enabled, action=action_type, match=match)

    status_code = int(action_data.get("status_code", 200))
    reason = str(action_data.get("reason", "OK"))
    headers_raw = action_data.get("headers", {})
    headers = dict(headers_raw) if isinstance(headers_raw, dict) else {}
    body_base64 = action_data.get("body_base64")
    if isinstance(body_base64, str):
        try:
            body = base64.b64decode(body_base64, validate=True)
        except ValueError:
            body = b""
    else:
        body = str(action_data.get("body", "")).encode("utf-8")
    template = StaticResponseTemplate(status_code=status_code, reason=reason, headers=headers, body=body)
    return PolicyRule(
        name=name,
        enabled=enabled,
        action=action_type,
        match=match,
        static_response=template,
    )



import json
import os
from pathlib import Path
import shlex
import shutil
import subprocess
import tempfile

from proxyscope.app.config.runtime import PolicyRule, parse_policy_rule, serialize_policy_rule


def edit_policy_rule_with_external_editor(rule: PolicyRule) -> tuple[bool, PolicyRule | None, str]:
    editor = _resolve_editor_command()
    if editor is None:
        return False, None, "No editor found. Set $EDITOR (or install nano/vim/vi)."

    payload = serialize_policy_rule(rule)
    try:
        with tempfile.TemporaryDirectory(prefix="tproxy-policy-") as tmp_dir:
            file_path = Path(tmp_dir) / "policy.json"
            file_path.write_text(json.dumps(payload, indent=2, ensure_ascii=True) + "\n", encoding="utf-8")

            print(f"[tproxy] Editing policy in {file_path.name} ({editor})")
            _run_editor(editor, file_path)

            edited_payload = json.loads(file_path.read_text(encoding="utf-8"))
            parsed = parse_policy_rule(edited_payload)
            if parsed is None:
                return False, None, "Edited policy is invalid JSON schema."
            if parsed.name != rule.name:
                parsed = PolicyRule(
                    name=rule.name,
                    enabled=parsed.enabled,
                    action=parsed.action,
                    match=parsed.match,
                    static_response=parsed.static_response,
                )
            return True, parsed, f"Policy updated: {rule.name}"
    except Exception as exc:  # noqa: BLE001
        return False, None, f"Policy edit failed ({exc})."


def _resolve_editor_command() -> str | None:
    env_editor = os.environ.get("EDITOR")
    if env_editor:
        parts = shlex.split(env_editor)
        if parts and shutil.which(parts[0]) is not None:
            return env_editor

    for candidate in ("nano", "vim", "vi"):
        if shutil.which(candidate) is not None:
            return candidate
    return None


def _run_editor(editor: str, file_path: Path) -> None:
    cmd = shlex.split(editor) + [str(file_path)]
    subprocess.run(cmd, check=True)



from dataclasses import dataclass
import curses
from typing import Literal

from proxyscope.app.runtime.journal import LoggedExchange

MainMode = Literal["requests", "request_detail"]
ActivePane = Literal["requests", "detail", "aux"]
DetailTab = Literal["request", "response"]


@dataclass(frozen=True)
class AuxPanelTabModel:
    key: str
    title: str
    items: list[str]
    cursor: int
    scroll: int
    empty_label: str = "<empty>"


@dataclass(frozen=True)
class RuntimeScreenModel:
    request_entries: list[LoggedExchange]
    request_cursor: int
    request_scroll: int
    request_detail_scroll: int
    response_detail_scroll: int
    main_mode: MainMode
    active_pane: ActivePane
    detail_tab: DetailTab
    aux_visible: bool
    aux_tabs: list[AuxPanelTabModel]
    aux_active_key: str
    command_buffer: str
    status_message: str
    config_text: str


@dataclass(frozen=True)
class RuntimeScreenResult:
    aux_scrolls: dict[str, int]
    request_scroll: int
    request_detail_scroll: int
    response_detail_scroll: int


class RuntimeScreenRenderer:
    def __init__(self) -> None:
        self._colors_initialized = False

    def draw(self, stdscr: "curses._CursesWindow", model: RuntimeScreenModel) -> RuntimeScreenResult:
        self._init_colors()
        stdscr.erase()
        rows, cols = stdscr.getmaxyx()
        if rows < 18 or cols < 90:
            _safe_addnstr(stdscr, 0, 0, "Terminal too small for runtime UI.", max(1, cols - 1), curses.A_BOLD)
            stdscr.refresh()
            return RuntimeScreenResult(
                aux_scrolls={tab.key: tab.scroll for tab in model.aux_tabs},
                request_scroll=model.request_scroll,
                request_detail_scroll=model.request_detail_scroll,
                response_detail_scroll=model.response_detail_scroll,
            )

        command_height = 6
        content_height = rows - command_height
        aux_width = _compute_aux_width(cols, model.aux_visible)
        main_width = cols - aux_width - (1 if aux_width > 0 else 0)

        request_scroll = model.request_scroll
        request_detail_scroll = model.request_detail_scroll
        response_detail_scroll = model.response_detail_scroll
        aux_scrolls = {tab.key: tab.scroll for tab in model.aux_tabs}

        if model.main_mode == "requests":
            request_scroll = _draw_requests_pane(
                stdscr,
                row=0,
                col=0,
                height=content_height,
                width=main_width,
                entries=model.request_entries,
                cursor=model.request_cursor,
                scroll=model.request_scroll,
                active=model.active_pane == "requests",
                selected_attr=self._selected_attr(),
            )
        else:
            request_area_width = _compute_request_list_width(
                total_main_width=main_width,
                aux_visible=model.aux_visible,
            )
            detail_area_width = main_width - request_area_width - 1
            request_scroll = _draw_requests_pane(
                stdscr,
                row=0,
                col=0,
                height=content_height,
                width=request_area_width,
                entries=model.request_entries,
                cursor=model.request_cursor,
                scroll=model.request_scroll,
                active=model.active_pane == "requests",
                selected_attr=self._selected_attr(),
            )
            selected_entry = model.request_entries[model.request_cursor] if model.request_entries else None
            request_detail_scroll, response_detail_scroll = _draw_detail_pane(
                stdscr,
                row=0,
                col=request_area_width + 1,
                height=content_height,
                width=detail_area_width,
                entry=selected_entry,
                active=model.active_pane == "detail",
                active_tab=model.detail_tab,
                request_scroll=model.request_detail_scroll,
                response_scroll=model.response_detail_scroll,
                selected_attr=self._selected_attr(),
            )

        if aux_width > 0:
            if model.main_mode == "request_detail":
                _safe_addch(stdscr, 0, main_width, "|")
                for y in range(1, content_height):
                    _safe_addch(stdscr, y, main_width, "|")
            aux_scrolls = _draw_aux_panel(
                stdscr,
                row=0,
                col=main_width + 1,
                height=content_height,
                width=aux_width,
                tabs=model.aux_tabs,
                active_key=model.aux_active_key,
                active=model.active_pane == "aux",
                selected_attr=self._selected_attr(),
            )

        _draw_command_area(
            stdscr,
            top=content_height,
            width=cols,
            command_buffer=model.command_buffer,
            config_text=model.config_text,
            status_message=model.status_message,
            title_attr=self._title_attr(),
            status_attr=self._status_attr(),
        )
        cursor_col = min(cols - 2, len(model.command_buffer) + 4)
        try:
            stdscr.move(content_height + 2, cursor_col)
        except curses.error:
            pass
        stdscr.refresh()

        return RuntimeScreenResult(
            aux_scrolls=aux_scrolls,
            request_scroll=request_scroll,
            request_detail_scroll=request_detail_scroll,
            response_detail_scroll=response_detail_scroll,
        )

    def _init_colors(self) -> None:
        if self._colors_initialized:
            return
        if not curses.has_colors():
            self._colors_initialized = True
            return
        try:
            curses.start_color()
            curses.use_default_colors()
            curses.init_pair(1, curses.COLOR_CYAN, -1)
            curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_CYAN)
            curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_WHITE)
        except curses.error:
            pass
        self._colors_initialized = True

    def _title_attr(self) -> int:
        if curses.has_colors():
            return curses.color_pair(1) | curses.A_BOLD
        return curses.A_BOLD

    def _status_attr(self) -> int:
        if curses.has_colors():
            return curses.color_pair(2) | curses.A_BOLD
        return curses.A_REVERSE

    def _selected_attr(self) -> int:
        if curses.has_colors():
            return curses.color_pair(3) | curses.A_BOLD
        return curses.A_REVERSE


def _draw_requests_pane(
    stdscr: "curses._CursesWindow",
    *,
    row: int,
    col: int,
    height: int,
    width: int,
    entries: list[LoggedExchange],
    cursor: int,
    scroll: int,
    active: bool,
    selected_attr: int,
) -> int:
    title = "REQUESTS"
    if active:
        title += " *"
    _draw_box(stdscr, row, col, height, width, title, active=active)
    inner_height = max(0, height - 2)
    inner_width = max(1, width - 2)
    normalized_scroll = _ensure_visible(cursor, scroll, inner_height)
    visible = entries[normalized_scroll: normalized_scroll + inner_height]
    for row_offset, entry in enumerate(visible):
        absolute_index = normalized_scroll + row_offset
        selected = absolute_index == cursor and active
        marker = ">" if selected else " "
        status = "---" if entry.response is None else str(entry.response.status_code)
        host = entry.target_host or "-"
        duration = "-" if entry.duration_ms is None else f"{entry.duration_ms:.1f}ms"
        row_text = (
            f"{marker}{entry.request_id:4d} {status:>3} "
            f"{entry.request.method:<7} {host:<20} {entry.request.path} [{duration}]"
        )
        attr = selected_attr if selected else 0
        _safe_addnstr(
            stdscr,
            row + 1 + row_offset,
            col + 1,
            _fit_text(row_text, inner_width),
            inner_width,
            attr,
        )
    return normalized_scroll


def _draw_detail_pane(
    stdscr: "curses._CursesWindow",
    *,
    row: int,
    col: int,
    height: int,
    width: int,
    entry: LoggedExchange | None,
    active: bool,
    active_tab: DetailTab,
    request_scroll: int,
    response_scroll: int,
    selected_attr: int,
) -> tuple[int, int]:
    title = f"DETAIL [{active_tab}]"
    if active:
        title += " *"
    _draw_box(stdscr, row, col, height, width, title, active=active)
    inner_height = max(0, height - 2)
    inner_width = max(1, width - 2)
    if entry is None:
        _safe_addnstr(stdscr, row + 1, col + 1, "No request selected.", inner_width, 0)
        return request_scroll, response_scroll

    request_tab = "[REQUEST]"
    response_tab = "[RESPONSE]"
    request_attr = selected_attr if active and active_tab == "request" else curses.A_BOLD
    response_attr = selected_attr if active and active_tab == "response" else curses.A_BOLD
    _safe_addnstr(stdscr, row + 1, col + 1, request_tab, min(len(request_tab), inner_width), request_attr)
    if inner_width > len(request_tab) + 1:
        _safe_addnstr(
            stdscr,
            row + 1,
            col + 1 + len(request_tab) + 1,
            response_tab,
            min(len(response_tab), inner_width - len(request_tab) - 1),
            response_attr,
        )

    content_row = row + 2
    content_height = max(0, inner_height - 1)
    if content_height == 0:
        return request_scroll, response_scroll

    request_lines = _build_request_lines(entry, inner_width)
    response_lines = _build_response_lines(entry, inner_width)
    request_max_scroll = max(0, len(request_lines) - content_height)
    response_max_scroll = max(0, len(response_lines) - content_height)
    request_scroll = min(max(0, request_scroll), request_max_scroll)
    response_scroll = min(max(0, response_scroll), response_max_scroll)

    lines = request_lines if active_tab == "request" else response_lines
    scroll = request_scroll if active_tab == "request" else response_scroll
    visible = lines[scroll: scroll + content_height]
    for row_offset, line in enumerate(visible):
        _safe_addnstr(stdscr, content_row + row_offset, col + 1, _fit_text(line, inner_width), inner_width, 0)
    return request_scroll, response_scroll


def _draw_aux_panel(
    stdscr: "curses._CursesWindow",
    *,
    row: int,
    col: int,
    height: int,
    width: int,
    tabs: list[AuxPanelTabModel],
    active_key: str,
    active: bool,
    selected_attr: int,
) -> dict[str, int]:
    title = "UTILITY [Shift+V hide]"
    if active:
        title += " *"
    _draw_box(stdscr, row, col, height, width, title, active=active)

    inner_width = max(1, width - 2)
    inner_height = max(0, height - 2)
    if not tabs:
        _safe_addnstr(stdscr, row + 1, col + 1, "<no tabs>", inner_width, 0)
        return {}

    tab_col = col + 1
    max_col = col + inner_width
    for tab in tabs:
        label = f"[{tab.title}]"
        remaining = max_col - tab_col + 1
        if remaining <= 1:
            break
        attr = selected_attr if active and tab.key == active_key else curses.A_BOLD
        _safe_addnstr(stdscr, row + 1, tab_col, label, remaining, attr)
        tab_col += len(label) + 1

    content_row = row + 2
    content_height = max(0, inner_height - 1)
    if content_height == 0:
        return {tab.key: tab.scroll for tab in tabs}

    selected_tab = next((tab for tab in tabs if tab.key == active_key), tabs[0])
    normalized = _ensure_visible(selected_tab.cursor, selected_tab.scroll, content_height)
    visible = selected_tab.items[normalized: normalized + content_height]
    if not visible:
        _safe_addnstr(
            stdscr,
            content_row,
            col + 1,
            _fit_text(selected_tab.empty_label, inner_width),
            inner_width,
            0,
        )
    for row_offset, item in enumerate(visible):
        absolute_index = normalized + row_offset
        selected = absolute_index == selected_tab.cursor and active
        marker = ">" if selected else " "
        row_text = f"{marker}{absolute_index + 1:3d}. {item}"
        attr = selected_attr if selected else 0
        _safe_addnstr(stdscr, content_row + row_offset, col + 1, _fit_text(row_text, inner_width), inner_width, attr)
    updated = {tab.key: tab.scroll for tab in tabs}
    updated[selected_tab.key] = normalized
    return updated


def _draw_command_area(
    stdscr: "curses._CursesWindow",
    *,
    top: int,
    width: int,
    command_buffer: str,
    config_text: str,
    status_message: str,
    title_attr: int,
    status_attr: int,
) -> None:
    for col_idx in range(width):
        _safe_addch(stdscr, top, col_idx, "-")
    _safe_addnstr(stdscr, top + 1, 1, "COMMAND", max(1, width - 2), title_attr)
    _safe_addnstr(stdscr, top + 2, 1, _safe_text(f"> {command_buffer}"), max(1, width - 2), 0)
    _safe_addnstr(stdscr, top + 3, 1, _fit_text(_safe_text(f"CONFIG: {config_text}"), max(1, width - 2)), max(1, width - 2), 0)
    _safe_addnstr(stdscr, top + 4, 1, _fit_text(_safe_text(status_message), max(1, width - 2)), max(1, width - 2), status_attr)


def _draw_box(
    stdscr: "curses._CursesWindow",
    row: int,
    col: int,
    height: int,
    width: int,
    title: str,
    *,
    active: bool,
) -> None:
    if height < 2 or width < 2:
        return
    bottom = row + height - 1
    right = col + width - 1
    _safe_addch(stdscr, row, col, "+")
    _safe_addch(stdscr, row, right, "+")
    _safe_addch(stdscr, bottom, col, "+")
    _safe_addch(stdscr, bottom, right, "+")
    for x in range(col + 1, right):
        _safe_addch(stdscr, row, x, "-")
        _safe_addch(stdscr, bottom, x, "-")
    for y in range(row + 1, bottom):
        _safe_addch(stdscr, y, col, "|")
        _safe_addch(stdscr, y, right, "|")
    attr = curses.A_BOLD | (curses.A_REVERSE if active else 0)
    _safe_addnstr(stdscr, row, col + 2, _fit_text(f"[ {title} ]", max(1, width - 4)), max(1, width - 4), attr)


def _compute_aux_width(total_cols: int, aux_visible: bool) -> int:
    if not aux_visible:
        return 0
    width = max(30, min(44, int(total_cols * 0.30)))
    if total_cols - width < 52:
        width = max(26, total_cols - 52)
    return max(0, width)


def _compute_request_list_width(*, total_main_width: int, aux_visible: bool) -> int:
    ratio = 0.30 if aux_visible else 0.34
    minimum = 26 if aux_visible else 30
    maximum = 44 if aux_visible else 52
    width = max(minimum, min(maximum, int(total_main_width * ratio)))
    if total_main_width - width < 42:
        width = max(minimum, total_main_width - 42)
    return max(minimum, width)


def _build_request_lines(entry: LoggedExchange, width: int) -> list[str]:
    lines = [
        f"REQUEST #{entry.request_id}",
        entry.request.start_line,
        f"Client: {entry.client_ip}",
        f"Target: {_target_label(entry)}",
        "",
        "Headers:",
    ]
    lines.extend(f"{name}: {value}" for name, value in entry.request.headers)
    lines.append("")
    lines.append("Body:")
    lines.extend(_body_lines(entry.request.body_preview))
    return _wrap_lines(lines, max(12, width))


def _build_response_lines(entry: LoggedExchange, width: int) -> list[str]:
    lines = ["RESPONSE"]
    if entry.response is None:
        lines.append("<pending>")
        return _wrap_lines(lines, max(12, width))
    lines.append(entry.response.start_line)
    lines.append("Headers:")
    lines.extend(f"{name}: {value}" for name, value in entry.response.headers)
    lines.append("")
    lines.append("Body:")
    lines.extend(_body_lines(entry.response.body_preview))
    return _wrap_lines(lines, max(12, width))


def _wrap_lines(lines: list[str], width: int) -> list[str]:
    wrapped: list[str] = []
    for line in lines:
        wrapped.extend(_wrap_line(line, width))
    return wrapped


def _target_label(entry: LoggedExchange) -> str:
    host = entry.target_host or "-"
    if entry.target_port is None:
        return host
    return f"{host}:{entry.target_port}"


def _body_lines(body_preview: str, *, max_lines: int = 20) -> list[str]:
    if not body_preview:
        return ["<empty>"]
    lines = body_preview.splitlines() or [body_preview]
    if len(lines) <= max_lines:
        return lines
    return lines[:max_lines] + ["...[truncated]"]


def _wrap_line(value: str, width: int) -> list[str]:
    text = _safe_text(value)
    if width <= 1:
        return [text[:1]]
    if text == "":
        return [""]
    chunks: list[str] = []
    current = text
    while len(current) > width:
        chunks.append(current[:width])
        current = current[width:]
    chunks.append(current)
    return chunks


def _fit_text(value: str, width: int) -> str:
    text = _safe_text(value)
    if width <= 0:
        return ""
    if len(text) <= width:
        return text
    if width <= 1:
        return text[:1]
    return text[: width - 1] + "~"


def _ensure_visible(cursor: int, scroll: int, height: int) -> int:
    if height <= 0:
        return 0
    if cursor < scroll:
        return cursor
    if cursor >= scroll + height:
        return cursor - height + 1
    return scroll


def _safe_addnstr(stdscr: "curses._CursesWindow", row: int, col: int, text: str, width: int, attr: int) -> None:
    try:
        stdscr.addnstr(row, col, text, width, attr)
    except curses.error:
        pass


def _safe_addch(stdscr: "curses._CursesWindow", row: int, col: int, value: str) -> None:
    try:
        stdscr.addch(row, col, value)
    except curses.error:
        pass


def _safe_text(value: str) -> str:
    return value.replace("\x00", "\\x00")



from dataclasses import dataclass
from urllib.parse import urlsplit

from proxyscope.app.config.models import PolicyRule


def normalize_whitelist_entry(value: str) -> str:
    raw_value = value.strip()
    if not raw_value:
        raise ValueError("Whitelist entry must not be empty.")

    if "://" in raw_value:
        parsed = urlsplit(raw_value)
        host = parsed.hostname
    elif "/" in raw_value:
        parsed = urlsplit(f"http://{raw_value}")
        host = parsed.hostname
    else:
        parsed = urlsplit(f"http://{raw_value}")
        host = parsed.hostname

    if not host:
        raise ValueError(f"Invalid whitelist entry: {value}")
    return host.lower()


def normalize_modification_url(value: str) -> str:
    raw_value = value.strip()
    if not raw_value:
        raise ValueError("Modification URL must not be empty.")

    if "://" not in raw_value:
        raw_value = f"http://{raw_value}"

    parsed = urlsplit(raw_value)
    if not parsed.hostname:
        raise ValueError(f"Invalid modification URL: {value}")

    scheme = parsed.scheme.lower() or "http"
    host = parsed.hostname.lower()
    port = parsed.port
    path = parsed.path or "/"
    normalized_path = path if path.startswith("/") else f"/{path}"
    netloc = f"{host}:{port}" if port is not None else host
    return f"{scheme}://{netloc}{normalized_path}"


def normalize_http_method(method: str | None) -> str:
    if method is None:
        raise ValueError("HTTP method must not be empty.")
    normalized = method.strip().upper()
    if not normalized:
        raise ValueError("HTTP method must not be empty.")
    return normalized


def request_url_candidates(url: str) -> tuple[str, ...]:
    normalized = normalize_modification_url(url)
    alternate = _alternate_scheme_url(normalized)
    if alternate is None:
        return (normalized,)
    return (normalized, alternate)


def policy_rule_matches_request(*, rule: PolicyRule, method: str, url_candidates: tuple[str, ...]) -> bool:
    methods = rule.match.methods
    if methods is not None and method not in methods:
        return False

    url_exact = rule.match.url_exact
    url_prefix = rule.match.url_prefix
    if url_exact is None and url_prefix is None:
        return True

    for candidate in url_candidates:
        if url_exact is not None and candidate == url_exact:
            return True
        if url_prefix is not None and candidate.startswith(url_prefix):
            return True
        # Host-wide fallback when configured with "/"
        candidate_parsed = _parse_normalized_modification_url(candidate)
        if url_exact is not None:
            entry = _parse_normalized_modification_url(url_exact)
            if entry.host == candidate_parsed.host and entry.port == candidate_parsed.port:
                if entry.path == "/" or candidate_parsed.path.startswith(entry.path):
                    return True
        if url_prefix is not None:
            entry = _parse_normalized_modification_url(url_prefix)
            if entry.host == candidate_parsed.host and entry.port == candidate_parsed.port:
                if entry.path == "/" or candidate_parsed.path.startswith(entry.path):
                    return True
    return False


def rule_matches_url(rule: PolicyRule, normalized_url: str) -> bool:
    target = rule.match.url_exact or rule.match.url_prefix
    if target is None:
        return False
    return target == normalized_url


def first_rule_method(rule: PolicyRule) -> str | None:
    methods = rule.match.methods
    if not methods:
        return None
    return methods[0]


def rule_method_display(rule: PolicyRule) -> str:
    method = first_rule_method(rule)
    return method or "*"


def rule_url_display(rule: PolicyRule) -> str:
    return rule.match.url_exact or rule.match.url_prefix or "*"


def policy_description(rule: PolicyRule) -> str:
    state = "on" if rule.enabled else "off"
    method = rule_method_display(rule)
    target = rule_url_display(rule)
    if rule.action == "open_editor":
        return f"{rule.name} [{state}] open_editor {method} {target}"
    if rule.action == "static_response" and rule.static_response is not None:
        return (
            f"{rule.name} [{state}] static_response {method} {target} "
            f"-> {rule.static_response.status_code} {rule.static_response.reason}"
        )
    return f"{rule.name} [{state}] {rule.action} {method} {target}"


def _alternate_scheme_url(normalized_url: str) -> str | None:
    parsed = urlsplit(normalized_url)
    scheme = parsed.scheme.lower()
    if scheme not in {"http", "https"}:
        return None
    alternate_scheme = "https" if scheme == "http" else "http"
    return f"{alternate_scheme}://{parsed.netloc}{parsed.path or '/'}"


@dataclass(frozen=True)
class _NormalizedModUrl:
    host: str
    port: int | None
    path: str


def _parse_normalized_modification_url(value: str) -> _NormalizedModUrl:
    parsed = urlsplit(value)
    return _NormalizedModUrl(
        host=(parsed.hostname or "").lower(),
        port=parsed.port,
        path=parsed.path or "/",
    )



import curses
from typing import Literal, Protocol

ActivePane = Literal["requests", "detail", "aux"]
MainMode = Literal["requests", "request_detail"]


class RuntimeInputHost(Protocol):
    _command_buffer: str
    _active_pane: ActivePane
    _aux_tab_key: str
    _main_mode: MainMode

    def _move_focus(self, direction: int) -> None: ...

    def _move_vertical(self, delta: int) -> None: ...

    def _page_vertical(self, direction: int) -> None: ...

    def _switch_to_request_list_mode(self) -> None: ...

    def _focus_aux_tab(self, tab_key: str) -> None: ...

    def _toggle_aux_visibility(self) -> None: ...

    def _add_selected_site_to_whitelist(self) -> None: ...

    def _remove_selected_site_from_whitelist(self) -> None: ...

    def _enable_selected_policy(self) -> None: ...

    def _disable_selected_policy(self) -> None: ...

    def _remove_selected_policy(self) -> None: ...

    def _edit_selected_policy(self, stdscr: "curses._CursesWindow") -> None: ...

    def _add_selected_request_to_modify_whitelist(self) -> None: ...

    def _edit_and_resend_selected_request(self, stdscr: "curses._CursesWindow") -> None: ...

    def _toggle_detail_tab(self) -> None: ...

    def _open_selected_request_detail(self) -> None: ...

    def execute_command(self, command: str) -> bool: ...


class RuntimeInputController:
    def __init__(self) -> None:
        self._key_btab = getattr(curses, "KEY_BTAB", -1)

    def handle_key(self, host: RuntimeInputHost, stdscr: "curses._CursesWindow", key: int) -> None:
        if key == -1:
            return

        if key == curses.KEY_LEFT:
            host._move_focus(-1)
            return

        if key == curses.KEY_RIGHT:
            host._move_focus(1)
            return

        if key == curses.KEY_UP:
            host._move_vertical(-1)
            return

        if key == curses.KEY_DOWN:
            host._move_vertical(1)
            return

        if key == curses.KEY_PPAGE:
            host._page_vertical(-1)
            return

        if key == curses.KEY_NPAGE:
            host._page_vertical(1)
            return

        if key == ord("B") and not host._command_buffer:
            host._switch_to_request_list_mode()
            return

        if key == ord("P") and not host._command_buffer:
            host._focus_aux_tab("policies")
            return

        if key == ord("S") and not host._command_buffer:
            host._focus_aux_tab("sites")
            return

        if key == ord("V") and not host._command_buffer:
            host._toggle_aux_visibility()
            return

        if key == ord("A") and not host._command_buffer and host._active_pane == "aux" and host._aux_tab_key == "sites":
            host._add_selected_site_to_whitelist()
            return

        if key == ord("D") and not host._command_buffer and host._active_pane == "aux" and host._aux_tab_key == "sites":
            host._remove_selected_site_from_whitelist()
            return

        if key == ord("E") and not host._command_buffer and host._active_pane == "aux" and host._aux_tab_key == "policies":
            host._enable_selected_policy()
            return

        if key == ord("D") and not host._command_buffer and host._active_pane == "aux" and host._aux_tab_key == "policies":
            host._disable_selected_policy()
            return

        if key == ord("X") and not host._command_buffer and host._active_pane == "aux" and host._aux_tab_key == "policies":
            host._remove_selected_policy()
            return

        if key == ord("I") and not host._command_buffer and host._active_pane == "aux" and host._aux_tab_key == "policies":
            host._edit_selected_policy(stdscr)
            return

        if key == ord("M") and not host._command_buffer:
            host._add_selected_request_to_modify_whitelist()
            return

        if key == ord("R") and not host._command_buffer:
            host._edit_and_resend_selected_request(stdscr)
            return

        if key == ord("T") and not host._command_buffer and host._main_mode == "request_detail":
            host._toggle_detail_tab()
            return

        if key in (9, self._key_btab) and not host._command_buffer and host._main_mode == "request_detail":
            host._toggle_detail_tab()
            return

        if key in (10, 13):
            if host._command_buffer:
                command = host._command_buffer
                host._command_buffer = ""
                host.execute_command(command)
                return
            host._open_selected_request_detail()
            return

        if key in (curses.KEY_BACKSPACE, 127, 8):
            host._command_buffer = host._command_buffer[:-1]
            return

        if 32 <= key <= 126:
            host._command_buffer += chr(key)



from argparse import ArgumentParser
import logging
import sys
import threading

from proxyscope.app.config.runtime import RuntimeConfig, set_runtime_config
from proxyscope.app.logging.setup import configure_logging
from proxyscope.app.logging.observability import set_runtime_observer
from proxyscope.app.runtime.journal import RequestJournal, set_request_journal
from proxyscope.app.editing.modifier import ResponseModifierService, set_response_modifier
from proxyscope.app.runtime.cli import RuntimeCLI
from proxyscope.proxy.server import ProxyHTTPServer, create_server

LOGGER = logging.getLogger("tproxy.app")


def _build_parser() -> ArgumentParser:
    parser = ArgumentParser(description="Minimal HTTP proxy learning server")
    parser.add_argument("--host", default="127.0.0.1", help="Host to bind")
    parser.add_argument("--port", default=8080, type=int, help="Port to bind")
    parser.add_argument("--config", default=None, help="Path to JSON runtime configuration file")
    parser.add_argument("--no-ui", action="store_true", help="Disable interactive runtime CLI UI")
    return parser


def main() -> None:
    args = _build_parser().parse_args()
    runtime_config = RuntimeConfig.load_from_file(args.config) if args.config else RuntimeConfig()
    request_journal = RequestJournal()
    response_modifier = ResponseModifierService()
    set_runtime_config(runtime_config)
    set_request_journal(request_journal)
    set_response_modifier(response_modifier)
    configure_logging(level=runtime_config.log_level)

    server = create_server(args.host, args.port)
    LOGGER.info("Listening on http://%s:%d", args.host, args.port)

    use_ui = not args.no_ui and sys.stdin.isatty() and sys.stdout.isatty()

    if not use_ui:
        response_modifier.set_interactive_enabled(False)
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            LOGGER.info("Shutdown requested")
        finally:
            server.server_close()
        return

    response_modifier.set_interactive_enabled(True)
    runtime_ui = RuntimeCLI(
        runtime_config=runtime_config,
        request_journal=request_journal,
        response_modifier=response_modifier,
        proxy_base_url=f"http://{args.host}:{args.port}",
    )
    runtime_ui.setFormatter(
        logging.Formatter("%(asctime)s | %(levelname)s | %(name)s | %(message)s", datefmt="%H:%M:%S")
    )

    root_logger = logging.getLogger()
    for handler in list(root_logger.handlers):
        root_logger.removeHandler(handler)
    root_logger.addHandler(runtime_ui)
    root_logger.setLevel(runtime_config.log_level)
    set_runtime_observer(runtime_ui)

    server_thread = threading.Thread(target=server.serve_forever, daemon=True)
    server_thread.start()

    def close_active_ssl_tunnels() -> None:
        if not isinstance(server, ProxyHTTPServer):
            return
        closed = server.close_all_active_tunnels()
        LOGGER.info("Closed %d active SSL tunnel(s) after cache setting change.", closed)

    try:
        runtime_ui.run(shutdown_server=server.shutdown, on_cache_toggle=close_active_ssl_tunnels)
    except KeyboardInterrupt:
        LOGGER.info("Shutdown requested")
    finally:
        server.shutdown()
        server.server_close()
        server_thread.join(timeout=2)
        set_runtime_observer(None)


if __name__ == "__main__":
    main()




from http.server import BaseHTTPRequestHandler

from proxyscope.proxy.forwarding import ForwardRequest, ForwardResponse


def map_incoming_request(handler: BaseHTTPRequestHandler) -> ForwardRequest:
    """Map http.server request data into the forwarding request model."""
    # `rfile` is a stream; read exactly Content-Length bytes to avoid partial reads
    # or blocking on methods that include a request body (for example POST/PUT).
    content_length = int(handler.headers.get("Content-Length", "0"))
    body = handler.rfile.read(content_length) if content_length > 0 else b""
    headers = {name: value for name, value in handler.headers.items()}
    return ForwardRequest(
        method=handler.command,
        path=handler.path,
        headers=headers,
        body=body,
    )


def write_forward_response(
    handler: BaseHTTPRequestHandler,
    response: ForwardResponse,
    *,
    send_body: bool,
) -> None:
    """Write forwarding response model data back to the HTTP client."""
    handler.send_response(response.status_code, response.reason)
    for name, value in response.headers.items():
        handler.send_header(name, value)
    handler.end_headers()

    if send_body:
        handler.wfile.write(response.body)



import json
import os
from pathlib import Path
import shlex
import shutil
import subprocess
import tempfile

import requests

from proxyscope.app.runtime.journal import LoggedExchange


def edit_and_resend_logged_request(
    entry: LoggedExchange,
    *,
    request_url: str,
    proxy_base_url: str | None,
) -> tuple[bool, str]:
    editor = _resolve_editor_command()
    if editor is None:
        return False, "No editor found. Set $EDITOR (or install nano/vim/vi)."

    payload = _build_edit_payload(entry, request_url=request_url)
    try:
        with tempfile.TemporaryDirectory(prefix="tproxy-replay-") as tmp_dir:
            file_path = Path(tmp_dir) / "request.json"
            file_path.write_text(json.dumps(payload, indent=2, ensure_ascii=True) + "\n", encoding="utf-8")

            print(f"[tproxy] Editing request replay payload in {file_path.name} ({editor})")
            _run_editor(editor, file_path)

            edited_payload = json.loads(file_path.read_text(encoding="utf-8"))
            replay = _parse_replay_payload(edited_payload)
            replay_headers = _sanitize_replay_headers(replay["headers"])

            proxies = None
            if proxy_base_url:
                proxies = {"http": proxy_base_url, "https": proxy_base_url}
            verify_tls = False if proxy_base_url else True

            response = requests.request(
                replay["method"],
                replay["url"],
                headers=replay_headers,
                data=replay["body"],
                timeout=60,
                allow_redirects=False,
                proxies=proxies,
                verify=verify_tls,
            )
            return True, f"Replayed {replay['method']} {replay['url']} -> {response.status_code} {response.reason}"
    except Exception as exc:  # noqa: BLE001
        return False, f"Replay failed ({exc})."


def _build_edit_payload(entry: LoggedExchange, *, request_url: str) -> dict[str, object]:
    headers = {name: value for name, value in entry.request.headers}
    return {
        "method": entry.request.method,
        "url": request_url,
        "headers": headers,
        # Body preview may be truncated in in-memory journal.
        "body_text": entry.request.body_preview,
    }


def _parse_replay_payload(payload: object) -> dict[str, object]:
    if not isinstance(payload, dict):
        raise ValueError("Replay payload must be a JSON object.")

    method_raw = payload.get("method", "GET")
    method = str(method_raw).strip().upper()
    if not method:
        raise ValueError("method must not be empty.")

    url_raw = payload.get("url", "")
    url = str(url_raw).strip()
    if not url:
        raise ValueError("url must not be empty.")
    if not (url.startswith("http://") or url.startswith("https://")):
        raise ValueError("url must start with http:// or https://")

    headers_raw = payload.get("headers", {})
    headers = dict(headers_raw) if isinstance(headers_raw, dict) else {}
    normalized_headers = {str(k): str(v) for k, v in headers.items()}

    body_text = str(payload.get("body_text", ""))
    body_bytes = body_text.encode("utf-8")

    return {
        "method": method,
        "url": url,
        "headers": normalized_headers,
        "body": body_bytes,
    }


def _sanitize_replay_headers(headers: dict[str, str]) -> dict[str, str]:
    sanitized = dict(headers)
    for name in (
        "Host",
        "Content-Length",
        "Connection",
        "Proxy-Connection",
        "Transfer-Encoding",
    ):
        _remove_header_case_insensitive(sanitized, name)
    return sanitized


def _remove_header_case_insensitive(headers: dict[str, str], header_name: str) -> None:
    target = header_name.lower()
    for key in list(headers.keys()):
        if key.lower() == target:
            del headers[key]


def _resolve_editor_command() -> str | None:
    env_editor = os.environ.get("EDITOR")
    if env_editor:
        parts = shlex.split(env_editor)
        if parts and shutil.which(parts[0]) is not None:
            return env_editor

    for candidate in ("nano", "vim", "vi"):
        if shutil.which(candidate) is not None:
            return candidate
    return None


def _run_editor(editor: str, file_path: Path) -> None:
    cmd = shlex.split(editor) + [str(file_path)]
    subprocess.run(cmd, check=True)



from dataclasses import dataclass
from urllib.parse import urlsplit

import requests

from proxyscope.app.config.runtime import is_cache_invalidation_enabled
from proxyscope.proxy.http1_request_rewriter import rewrite_cache_invalidation_headers


@dataclass(frozen=True)
class ForwardRequest:
    method: str
    path: str
    headers: dict[str, str]
    body: bytes = b""


@dataclass(frozen=True)
class ForwardResponse:
    status_code: int
    reason: str
    headers: dict[str, str]
    body: bytes


class UpstreamForwarder:
    """Forward a request to its dynamic upstream target."""

    def forward(self, request: ForwardRequest) -> ForwardResponse:
        url = resolve_target_url(request)
        headers = prepare_forward_headers(request.headers)
        response = requests.request(request.method,
                                    url,
                                    headers=headers,
                                    data=request.body,
                                    timeout=60)
        return ForwardResponse(response.status_code,
                               response.reason,
                               dict(response.headers),
                               response.content)


def resolve_target_url(request: ForwardRequest) -> str:
    if request.path.startswith(("http://", "https://")):
        parsed = urlsplit(request.path)
        if not parsed.netloc:
            raise ValueError("Invalid absolute-form request target.")
        return request.path

    host = request.headers.get("Host")
    if not host:
        raise ValueError("Missing Host header for dynamic forward proxy request.")

    normalized_path = request.path if request.path.startswith("/") else f"/{request.path}"
    return f"http://{host}{normalized_path}"


def _apply_cache_invalidation_headers(headers: dict[str, str]) -> None:
    """
    Force upstream cache re-fetch by removing conditional validators and
    setting explicit no-cache request directives.
    """
    rewritten = rewrite_cache_invalidation_headers(headers)
    headers.clear()
    headers.update(rewritten)


def prepare_forward_headers(headers: dict[str, str]) -> dict[str, str]:
    prepared = dict(headers)
    if is_cache_invalidation_enabled():
        _apply_cache_invalidation_headers(prepared)
    return prepared


def prepare_forward_request(request: ForwardRequest) -> ForwardRequest:
    return request



from threading import Lock


class TunnelConnectionRegistry:
    """
    Thread-safe registry for currently active CONNECT/TLS tunnel sockets.
    """

    def __init__(self) -> None:
        self._lock = Lock()
        self._sockets: set[object] = set()

    def register(self, sock: object) -> None:
        with self._lock:
            self._sockets.add(sock)

    def unregister(self, sock: object) -> None:
        with self._lock:
            self._sockets.discard(sock)

    def close_all(self) -> int:
        with self._lock:
            sockets = list(self._sockets)
            self._sockets.clear()

        closed = 0
        for sock in sockets:
            try:
                sock.close()  # type: ignore[attr-defined]
                closed += 1
            except OSError:
                pass
        return closed



from typing import Protocol

from proxyscope.proxy.forwarding import ForwardRequest, ForwardResponse


class Forwarder(Protocol):
    def forward(self, request: ForwardRequest) -> ForwardResponse:
        ...



from dataclasses import dataclass
import ipaddress
from pathlib import Path
import re
import subprocess


class MitmCertificateError(Exception):
    """Raised when CA or leaf certificate generation fails."""


_SAFE_HOST_RE = re.compile(r"[^A-Za-z0-9_.-]+")


@dataclass(frozen=True)
class MitmCertificateAuthority:
    ca_cert_path: Path
    ca_key_path: Path
    hosts_dir: Path

    def is_ready(self) -> bool:
        return self.ca_cert_path.exists() and self.ca_key_path.exists()

    def ensure_ca_material(self) -> bool:
        """
        Ensure CA key/cert material exists and return True when a new CA was created.
        """
        try:
            self.hosts_dir.mkdir(parents=True, exist_ok=True)
            if self.is_ready():
                self._write_browser_import_copy()
                return False

            self.ca_cert_path.parent.mkdir(parents=True, exist_ok=True)
            # If only one file exists, discard it and create a fresh key/cert pair.
            if self.ca_cert_path.exists() != self.ca_key_path.exists():
                self.ca_cert_path.unlink(missing_ok=True)
                self.ca_key_path.unlink(missing_ok=True)

            config_path = self.ca_cert_path.parent / "mitm-ca.cnf"
            config_path.write_text(
                "\n".join(
                    [
                        "[ req ]",
                        "distinguished_name = dn",
                        "x509_extensions = v3_ca",
                        "prompt = no",
                        "",
                        "[ dn ]",
                        "CN = proxyscope Local MITM CA",
                        "",
                        "[ v3_ca ]",
                        "subjectKeyIdentifier = hash",
                        "authorityKeyIdentifier = keyid:always,issuer",
                        "basicConstraints = critical, CA:true",
                        "keyUsage = critical, keyCertSign, cRLSign",
                        "",
                    ]
                ),
                encoding="utf-8",
            )

            self._run_openssl(
                [
                    "req",
                    "-x509",
                    "-newkey",
                    "rsa:4096",
                    "-sha256",
                    "-days",
                    "3650",
                    "-nodes",
                    "-config",
                    str(config_path),
                    "-extensions",
                    "v3_ca",
                    "-keyout",
                    str(self.ca_key_path),
                    "-out",
                    str(self.ca_cert_path),
                ]
            )
            if not self.is_ready():
                raise MitmCertificateError(
                    f"Failed to create CA files: cert={self.ca_cert_path} key={self.ca_key_path}"
                )
            self._write_browser_import_copy()
            return True
        except OSError as exc:
            raise MitmCertificateError(f"Filesystem error while preparing CA material: {exc}") from exc

    def issue_host_certificate(self, host: str) -> tuple[Path, Path]:
        if not self.is_ready():
            raise MitmCertificateError(
                f"Missing CA files: cert={self.ca_cert_path} key={self.ca_key_path}"
            )

        self.hosts_dir.mkdir(parents=True, exist_ok=True)
        safe_host = _SAFE_HOST_RE.sub("_", host)
        cert_path = self.hosts_dir / f"{safe_host}.cert.pem"
        key_path = self.hosts_dir / f"{safe_host}.key.pem"
        csr_path = self.hosts_dir / f"{safe_host}.csr.pem"
        ext_path = self.hosts_dir / f"{safe_host}.ext.cnf"
        serial_path = self.hosts_dir / "mitm-ca.srl"

        if cert_path.exists() and key_path.exists():
            if self._cert_matches_private_key(cert_path=cert_path, key_path=key_path):
                return cert_path, key_path
            cert_path.unlink(missing_ok=True)
            key_path.unlink(missing_ok=True)

        san_prefix = "IP" if _is_ip_address(host) else "DNS"
        ext_path.write_text(
            "\n".join(
                [
                    "basicConstraints=CA:FALSE",
                    "keyUsage=critical,digitalSignature,keyEncipherment",
                    "extendedKeyUsage=serverAuth",
                    "subjectKeyIdentifier=hash",
                    "authorityKeyIdentifier=keyid,issuer",
                    f"subjectAltName={san_prefix}:{host}",
                    "",
                ]
            ),
            encoding="utf-8",
        )

        self._run_openssl(
            [
                "req",
                "-new",
                "-newkey",
                "rsa:2048",
                "-nodes",
                "-keyout",
                str(key_path),
                "-out",
                str(csr_path),
                "-subj",
                f"/CN={host}",
            ]
        )
        self._run_openssl(
            [
                "x509",
                "-req",
                "-in",
                str(csr_path),
                "-CA",
                str(self.ca_cert_path),
                "-CAkey",
                str(self.ca_key_path),
                "-CAserial",
                str(serial_path),
                "-CAcreateserial",
                "-out",
                str(cert_path),
                "-days",
                "825",
                "-sha256",
                "-extfile",
                str(ext_path),
            ]
        )

        # Keep generated cert/key; cleanup intermediate files.
        if csr_path.exists():
            csr_path.unlink()
        if ext_path.exists():
            ext_path.unlink()

        return cert_path, key_path

    def _cert_matches_private_key(self, *, cert_path: Path, key_path: Path) -> bool:
        """
        Return True only if certificate and private key expose the same public key.
        """
        cert_pub = self._read_openssl_output(["x509", "-in", str(cert_path), "-pubkey", "-noout"])
        key_pub = self._read_openssl_output(["pkey", "-in", str(key_path), "-pubout"])
        if cert_pub is None or key_pub is None:
            return False
        return cert_pub.strip() == key_pub.strip()

    def _read_openssl_output(self, args: list[str]) -> str | None:
        command = ["openssl", *args]
        try:
            completed = subprocess.run(command, check=True, capture_output=True, text=True)
        except (FileNotFoundError, subprocess.CalledProcessError):
            return None
        return completed.stdout

    def _run_openssl(self, args: list[str]) -> None:
        command = ["openssl", *args]
        try:
            subprocess.run(command, check=True, capture_output=True, text=True)
        except FileNotFoundError as exc:
            raise MitmCertificateError("openssl is required but was not found in PATH.") from exc
        except subprocess.CalledProcessError as exc:
            stderr = exc.stderr.strip()
            raise MitmCertificateError(f"openssl failed: {' '.join(command)} | {stderr}") from exc

    def _write_browser_import_copy(self) -> None:
        browser_import_path = _browser_import_ca_path(self.ca_cert_path)
        browser_import_path.write_bytes(self.ca_cert_path.read_bytes())


def default_ca() -> MitmCertificateAuthority:
    certs_root = Path("certs")
    return MitmCertificateAuthority(
        ca_cert_path=certs_root / "ca" / "mitm-ca.cert.pem",
        ca_key_path=certs_root / "ca" / "mitm-ca.key.pem",
        hosts_dir=certs_root / "hosts",
    )


def _is_ip_address(host: str) -> bool:
    try:
        ipaddress.ip_address(host)
        return True
    except ValueError:
        return False


def _browser_import_ca_path(ca_cert_path: Path) -> Path:
    cert_suffix = ".cert.pem"
    cert_name = ca_cert_path.name
    if cert_name.endswith(cert_suffix):
        return ca_cert_path.with_name(f"{cert_name[: -len(cert_suffix)]}.crt")
    return ca_cert_path.with_suffix(".crt")



import logging


def configure_logging(level: int = logging.INFO) -> None:
    """Configure basic console logging for local development."""
    logging.basicConfig(
        level=level,
        format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )




from collections.abc import Callable

from proxyscope.app.config.runtime import get_static_response_template_for_request
from proxyscope.app.editing.modifier import ResponseModifierService
from proxyscope.proxy.forwarding import ForwardResponse


class HTTP1ResponseModifierRewriter:
    """
    Buffer and parse HTTP/1.x responses from a TLS stream, and allow
    interactive modification before forwarding bytes to the client.
    """

    def __init__(
        self,
        *,
        response_modifier: ResponseModifierService,
        acquire_request_meta: Callable[[], tuple[str, str] | None],
    ) -> None:
        self._response_modifier = response_modifier
        self._acquire_request_meta = acquire_request_meta
        self._buffer = bytearray()
        self._passthrough = False
        self._active_request_meta: tuple[str, str] | None = None

    def feed(self, data: bytes) -> bytes:
        if self._passthrough:
            return data

        self._buffer.extend(data)
        out = bytearray()

        while True:
            header_end = self._buffer.find(b"\r\n\r\n")
            if header_end < 0:
                break

            raw_header_block = bytes(self._buffer[:header_end])
            lines = raw_header_block.decode("iso-8859-1", errors="replace").split("\r\n")
            if not lines:
                self._passthrough = True
                out.extend(self._flush_buffer())
                break

            start_line = lines[0].strip()
            headers: dict[str, str] = {}
            for line in lines[1:]:
                if not line or ":" not in line:
                    continue
                name, value = line.split(":", 1)
                headers[name.strip()] = value.strip()

            if self._active_request_meta is None:
                self._active_request_meta = self._acquire_request_meta()
            request_meta = self._active_request_meta
            method = request_meta[0] if request_meta is not None else "GET"
            request_url = request_meta[1] if request_meta is not None else "https://unknown/"

            consumed = header_end + 4
            transfer_encoding = _header_value(headers, "transfer-encoding").lower()
            content_length_value = _header_value(headers, "content-length")

            raw_body = b""
            decoded_body = b""

            if _response_has_no_body(method, start_line):
                decoded_body = b""
                raw_body = b""
            elif "chunked" in transfer_encoding:
                parsed = _try_consume_chunked(bytes(self._buffer[consumed:]))
                if parsed is None:
                    break
                raw_chunked, decoded = parsed
                raw_body = raw_chunked
                decoded_body = decoded
                consumed += len(raw_chunked)
            elif content_length_value:
                try:
                    content_length = max(0, int(content_length_value))
                except ValueError:
                    self._passthrough = True
                    out.extend(self._flush_buffer())
                    break
                if len(self._buffer) < consumed + content_length:
                    break
                raw_body = bytes(self._buffer[consumed : consumed + content_length])
                decoded_body = raw_body
                consumed += content_length
            else:
                # Unknown body framing (usually close-delimited). Do not rewrite.
                self._passthrough = True
                out.extend(self._flush_buffer())
                break

            del self._buffer[:consumed]

            status_code, reason, version = _parse_response_start_line(start_line)
            if _is_interim_response(status_code):
                out.extend(raw_header_block)
                out.extend(b"\r\n\r\n")
                out.extend(raw_body)
                continue

            original_response = ForwardResponse(
                status_code=status_code,
                reason=reason,
                headers=headers,
                body=decoded_body,
            )
            static_template = get_static_response_template_for_request(method=method, url=request_url)
            if static_template is not None:
                original_response = ForwardResponse(
                    status_code=static_template.status_code,
                    reason=static_template.reason,
                    headers=dict(static_template.headers),
                    body=static_template.body,
                )
            edited_response = self._response_modifier.maybe_modify_response(
                request_url=request_url,
                method=method,
                response=original_response,
            )

            if edited_response is original_response and static_template is None:
                out.extend(raw_header_block)
                out.extend(b"\r\n\r\n")
                out.extend(raw_body)
                self._active_request_meta = None
                continue

            out.extend(_build_response_bytes(version=version, response=edited_response))
            self._active_request_meta = None

        return bytes(out)

    def flush(self) -> bytes:
        return self._flush_buffer()

    def _flush_buffer(self) -> bytes:
        if not self._buffer:
            return b""
        payload = bytes(self._buffer)
        self._buffer.clear()
        return payload


def _response_has_no_body(method: str, start_line: str) -> bool:
    code, _reason, _version = _parse_response_start_line(start_line)
    if method.upper() == "HEAD":
        return True
    if 100 <= code < 200:
        return True
    return code in {204, 304}


def _parse_response_start_line(start_line: str) -> tuple[int, str, str]:
    parts = start_line.split(" ", 2)
    version = parts[0] if parts else "HTTP/1.1"
    if len(parts) >= 2:
        try:
            code = int(parts[1])
        except ValueError:
            code = 200
    else:
        code = 200
    reason = parts[2] if len(parts) >= 3 else ""
    return code, reason, version


def _is_interim_response(status_code: int) -> bool:
    return 100 <= status_code < 200 and status_code != 101


def _build_response_bytes(*, version: str, response: ForwardResponse) -> bytes:
    headers = dict(response.headers)
    _remove_header_case_insensitive(headers, "Transfer-Encoding")
    headers["Content-Length"] = str(len(response.body))
    start_line = f"{version} {response.status_code} {response.reason}".rstrip()
    lines = [start_line]
    for name, value in headers.items():
        lines.append(f"{name}: {value}")
    return "\r\n".join(lines).encode("iso-8859-1", errors="replace") + b"\r\n\r\n" + response.body


def _remove_header_case_insensitive(headers: dict[str, str], header_name: str) -> None:
    target = header_name.lower()
    for key in list(headers.keys()):
        if key.lower() == target:
            del headers[key]


def _header_value(headers: dict[str, str], name: str) -> str:
    target = name.lower()
    for key, value in headers.items():
        if key.lower() == target:
            return value
    return ""


def _try_consume_chunked(data: bytes) -> tuple[bytes, bytes] | None:
    idx = 0
    decoded = bytearray()
    data_len = len(data)

    while True:
        line_end = data.find(b"\r\n", idx)
        if line_end < 0:
            return None
        line = data[idx:line_end]
        try:
            chunk_size = int(line.split(b";", 1)[0], 16)
        except ValueError:
            return None
        idx = line_end + 2

        if chunk_size == 0:
            trailer_end = data.find(b"\r\n\r\n", idx)
            if trailer_end >= 0:
                idx = trailer_end + 4
                return data[:idx], bytes(decoded)
            if idx + 2 <= data_len and data[idx:idx + 2] == b"\r\n":
                idx += 2
                return data[:idx], bytes(decoded)
            return None

        if idx + chunk_size + 2 > data_len:
            return None
        decoded.extend(data[idx: idx + chunk_size])
        idx = idx + chunk_size
        if data[idx:idx + 2] != b"\r\n":
            return None
        idx += 2



from collections import deque
from dataclasses import dataclass
import socket
import ssl
import threading
import time

from proxyscope.app.config.runtime import is_cache_invalidation_enabled
from proxyscope.mitm.certificates import MitmCertificateAuthority
from proxyscope.app.editing.modifier import get_response_modifier
from proxyscope.proxy.connect_tunnel import ConnectTarget, ConnectUpstreamConnectionError, ConnectUpstreamTimeoutError
from proxyscope.proxy.http1_request_rewriter import HTTP1RequestHeaderRewriter
from proxyscope.proxy.http1_response_modifier_rewriter import HTTP1ResponseModifierRewriter
from proxyscope.proxy.http1_sniffer import HTTP1MessageSniffer
from proxyscope.app.logging.request_response import log_mitm_http_request, log_mitm_http_response


@dataclass(frozen=True)
class MitmTLSInterceptor:
    certificate_authority: MitmCertificateAuthority

    def intercept(self, *, client_socket: socket.socket, target: ConnectTarget, timeout_s: float = 30.0) -> bool:
        """
        Perform MITM TLS interception for a CONNECT tunnel.

        Flow:
        1) connect proxy->upstream and establish TLS (as a client)
        2) send 200 Connection Established to the browser
        3) establish TLS browser->proxy (as a server with forged host cert)
        4) relay decrypted bytes between both TLS sessions

        Returns:
        - True if tunnel handshake succeeded and relay ran.
        - False if client rejected proxy certificate during TLS handshake.
        """
        try:
            upstream_tcp = socket.create_connection((target.host, target.port), timeout=timeout_s)
        except socket.timeout as exc:
            raise ConnectUpstreamTimeoutError(
                f"Timed out connecting to upstream target {target.host}:{target.port}."
            ) from exc
        except OSError as exc:
            raise ConnectUpstreamConnectionError(
                f"Failed to connect to upstream target {target.host}:{target.port}: {exc}"
            ) from exc

        with upstream_tcp:
            upstream_ctx = ssl.create_default_context()
            # Prefer HTTP/1.1 in MITM mode to keep decrypted stream logging practical.
            upstream_ctx.set_alpn_protocols(["http/1.1"])
            try:
                upstream_tls = upstream_ctx.wrap_socket(upstream_tcp, server_hostname=target.host)
            except ssl.SSLError as exc:
                raise ConnectUpstreamConnectionError(
                    f"Failed TLS handshake with upstream target {target.host}:{target.port}: {exc}"
                ) from exc

            with upstream_tls:
                # Tell the client to start TLS over this proxy connection.
                client_socket.sendall(b"HTTP/1.1 200 Connection Established\r\n\r\n")

                cert_path, key_path = self.certificate_authority.issue_host_certificate(target.host)
                client_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
                client_ctx.set_alpn_protocols(["http/1.1"])
                client_ctx.load_cert_chain(certfile=str(cert_path), keyfile=str(key_path))
                try:
                    client_tls = client_ctx.wrap_socket(client_socket, server_side=True)
                except ssl.SSLError:
                    # Client rejected the forged MITM certificate (usually trust issue).
                    # At this point HTTP 200 was already sent, so we just stop tunneling.
                    return False

                with client_tls:
                    try:
                        client_ip = client_socket.getpeername()[0]
                    except OSError:
                        client_ip = "unknown"

                    pending_request_ids: deque[int | None] = deque()
                    pending_request_meta: deque[tuple[str, str]] = deque()
                    pending_lock = threading.Lock()

                    def on_request(start_line: str, headers: dict[str, str], body: bytes) -> None:
                        request_id = log_mitm_http_request(
                            client_ip=client_ip,
                            target_host=target.host,
                            target_port=target.port,
                            start_line=start_line,
                            headers=headers,
                            body=body,
                        )
                        method, path = _parse_request_start_line(start_line)
                        request_url = _build_https_request_url(host=target.host, port=target.port, path=path)
                        with pending_lock:
                            pending_request_ids.append(request_id)
                            pending_request_meta.append((method, request_url))

                    def on_response(start_line: str, headers: dict[str, str], body: bytes) -> None:
                        with pending_lock:
                            request_id = pending_request_ids.popleft() if pending_request_ids else None
                        log_mitm_http_response(
                            client_ip=client_ip,
                            target_host=target.host,
                            target_port=target.port,
                            start_line=start_line,
                            headers=headers,
                            body=body,
                            request_id=request_id,
                        )

                    def acquire_request_meta() -> tuple[str, str] | None:
                        with pending_lock:
                            if not pending_request_meta:
                                return None
                            return pending_request_meta.popleft()

                    request_sniffer = HTTP1MessageSniffer(
                        on_request
                    )
                    response_sniffer = HTTP1MessageSniffer(
                        on_response
                    )
                    response_rewriter = HTTP1ResponseModifierRewriter(
                        response_modifier=get_response_modifier(),
                        acquire_request_meta=acquire_request_meta,
                    )
                    _relay_tls_bidirectional(
                        client_tls,
                        upstream_tls,
                        timeout_s=timeout_s,
                        request_sniffer=request_sniffer,
                        response_sniffer=response_sniffer,
                        response_rewriter=response_rewriter,
                        rewrite_client_requests=is_cache_invalidation_enabled(),
                    )
                return True


def _relay_tls_bidirectional(
    client_tls: ssl.SSLSocket,
    upstream_tls: ssl.SSLSocket,
    *,
    timeout_s: float,
    request_sniffer: HTTP1MessageSniffer,
    response_sniffer: HTTP1MessageSniffer,
    response_rewriter: HTTP1ResponseModifierRewriter,
    rewrite_client_requests: bool,
) -> None:
    """
    Relay bytes in both directions using two threads.
    """
    stop_event = threading.Event()
    relay_error: list[Exception] = []
    relay_lock = threading.Lock()
    idle_deadline = time.monotonic() + timeout_s
    idle_lock = threading.Lock()

    def pump(
        source: ssl.SSLSocket,
        destination: ssl.SSLSocket,
        sniffer: HTTP1MessageSniffer,
        *,
        rewriter: HTTP1RequestHeaderRewriter | None = None,
        response_modifier_rewriter: HTTP1ResponseModifierRewriter | None = None,
    ) -> None:
        nonlocal idle_deadline
        source.settimeout(timeout_s)
        destination.settimeout(timeout_s)
        try:
            while not stop_event.is_set():
                try:
                    data = source.recv(65536)
                except socket.timeout:
                    # Idle socket timeouts are expected for keep-alive HTTPS connections.
                    # Close gracefully only if both directions stayed idle beyond timeout_s.
                    with idle_lock:
                        if time.monotonic() >= idle_deadline:
                            return
                    continue

                if not data:
                    if rewriter is not None:
                        flushed = rewriter.flush()
                        if flushed:
                            sniffer.feed(flushed)
                            destination.sendall(flushed)
                    if response_modifier_rewriter is not None:
                        flushed = response_modifier_rewriter.flush()
                        if flushed:
                            sniffer.feed(flushed)
                            destination.sendall(flushed)
                    return
                with idle_lock:
                    idle_deadline = time.monotonic() + timeout_s
                payload = data
                if rewriter is not None:
                    payload = rewriter.feed(data)
                if response_modifier_rewriter is not None:
                    payload = response_modifier_rewriter.feed(payload)
                if not payload:
                    continue
                sniffer.feed(payload)
                destination.sendall(payload)
        except Exception as exc:  # noqa: BLE001
            with relay_lock:
                relay_error.append(exc)
        finally:
            stop_event.set()
            try:
                destination.shutdown(socket.SHUT_WR)
            except OSError:
                pass

    request_rewriter = HTTP1RequestHeaderRewriter() if rewrite_client_requests else None
    t1 = threading.Thread(
        target=pump,
        args=(client_tls, upstream_tls, request_sniffer),
        kwargs={"rewriter": request_rewriter},
        daemon=True,
    )
    t2 = threading.Thread(
        target=pump,
        args=(upstream_tls, client_tls, response_sniffer),
        kwargs={"response_modifier_rewriter": response_rewriter},
        daemon=True,
    )
    t1.start()
    t2.start()
    t1.join()
    t2.join()

    if relay_error:
        first_error = relay_error[0]
        if isinstance(first_error, ConnectUpstreamTimeoutError):
            raise first_error
        # Ignore clean EOF-like SSL errors; raise only obvious transport issues.
        if isinstance(first_error, (ssl.SSLError, OSError)):
            return
        raise first_error


def _parse_request_start_line(start_line: str) -> tuple[str, str]:
    parts = start_line.split(" ", 2)
    if len(parts) >= 2:
        return parts[0], parts[1]
    return "GET", "/"


def _build_https_request_url(*, host: str, port: int, path: str) -> str:
    if path.startswith("http://") or path.startswith("https://"):
        return path
    normalized_path = path if path.startswith("/") else f"/{path}"
    if port == 443:
        return f"https://{host}{normalized_path}"
    return f"https://{host}:{port}{normalized_path}"
